# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..serialize import DataTypeField
from ..operands import Operand, TileableOperandMixin, HasInput, ShuffleProxy, MapReduceOperand, \
    Fuse, OutputType
from ..utils import calc_nsplits


class TensorOperandMixin(TileableOperandMixin):
    __slots__ = ()
    _op_module_ = 'tensor'
    _output_type_ = OutputType.tensor

    def new_tensors(self, inputs, shape=None, dtype=None, order=None, chunks=None, nsplits=None,
                    output_limit=None, kws=None, **kw):
        return self.new_tileables(inputs, shape=shape, chunks=chunks, nsplits=nsplits,
                                  output_limit=output_limit, kws=kws, dtype=dtype, order=order, **kw)

    def new_tensor(self, inputs, shape, dtype=None, order=None, **kw):
        if getattr(self, 'output_limit') != 1:
            raise TypeError('cannot new tensor with more than 1 outputs')
        return self.new_tensors(inputs, shape=shape, dtype=dtype, order=order, **kw)[0]

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        from .merge.concatenate import TensorConcatenate

        tensor = tileable
        assert not tensor.is_coarse()

        op = TensorConcatenate(dtype=tensor.dtype)
        chunk = TensorConcatenate(dtype=tensor.dtype).new_chunk(
            tensor.chunks, shape=tensor.shape, index=(0,) * tileable.ndim)
        return op.new_tensor([tensor], tensor.shape, chunks=[chunk],
                             nsplits=tuple((s,) for s in tensor.shape))

    @classmethod
    def create_tileable_from_chunks(cls, chunks, inputs=None, **kw):
        chunk_idx_to_shape = {c.index: c.shape for c in chunks}
        nsplits = calc_nsplits(chunk_idx_to_shape)
        shape = tuple(sum(ns) for ns in nsplits)
        op = chunks[0].op.copy().reset_key()
        return op.new_tensor(inputs, shape=shape, chunks=chunks,
                             nsplits=nsplits, dtype=chunks[0].dtype, **kw)

    def get_fuse_op_cls(self, _):
        from .fuse import TensorFuseChunk

        return TensorFuseChunk


class TensorOperand(Operand):
    _output_type_ = OutputType.tensor
    _dtype = DataTypeField('dtype')

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)


class TensorHasInput(HasInput):
    _output_type_ = OutputType.tensor
    _dtype = DataTypeField('dtype')

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)


class TensorShuffleProxy(ShuffleProxy, TensorOperandMixin):
    _output_type_ = OutputType.tensor
    _dtype = DataTypeField('dtype')

    def __init__(self, dtype=None, **kwargs):
        kwargs['_dtype'] = kwargs.get('_dtype', dtype)
        super().__init__(**kwargs)

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)

    @classmethod
    def execute(cls, ctx, op):
        pass


class TensorMapReduceOperand(MapReduceOperand):
    _output_type_ = OutputType.tensor
    _dtype = DataTypeField('dtype')

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)


class TensorFuse(Fuse):
    _output_type_ = OutputType.tensor
    _dtype = DataTypeField('dtype')

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import random
import time
import uuid

from .actors import new_client, ActorNotExist
from .errors import GraphNotExists
from .scheduler import SessionActor, GraphActor, GraphMetaActor, ResourceActor, \
    SessionManagerActor, ChunkMetaClient
from .scheduler.node_info import NodeInfoActor
from .scheduler.utils import SchedulerClusterInfoActor
from .worker.transfer import ResultSenderActor, ReceiverManagerActor
from .tensor.utils import slice_split
from .serialize import dataserializer
from .utils import tokenize, merge_chunks, arrow_array_to_objects

logger = logging.getLogger(__name__)


class MarsAPI(object):
    def __init__(self, scheduler_ip):
        self.__schedulers_cache = None
        self._session_manager = None
        self.actor_client = new_client()
        self.cluster_info = self.actor_client.actor_ref(
            SchedulerClusterInfoActor.default_uid(), address=scheduler_ip)
        self.chunk_meta_client = ChunkMetaClient(self.actor_client, self.cluster_info)

    @property
    def session_manager(self):
        if self._session_manager is None:
            self._session_manager = self.get_actor_ref(SessionManagerActor.default_uid())
        return self._session_manager

    def get_schedulers(self):
        if not self.__schedulers_cache:
            self.__schedulers_cache = self.cluster_info.get_schedulers()
        return self.__schedulers_cache

    def get_scheduler(self, uid):
        schedulers = self.get_schedulers()
        if len(schedulers) == 1:
            return schedulers[0]
        else:
            return self.cluster_info.get_scheduler(uid)

    def get_actor_ref(self, uid):
        return self.actor_client.actor_ref(uid, address=self.get_scheduler(uid))

    def get_graph_meta_ref(self, session_id, graph_key):
        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_meta_uid = GraphMetaActor.gen_uid(session_id, graph_key)
        graph_addr = self.get_scheduler(graph_uid)
        return self.actor_client.actor_ref(graph_meta_uid, address=graph_addr)

    def get_schedulers_info(self):
        schedulers = self.get_schedulers()
        infos = dict()
        for scheduler in schedulers:
            info_ref = self.actor_client.actor_ref(NodeInfoActor.default_uid(), address=scheduler)
            infos[scheduler] = info_ref.get_info()
        return infos

    def count_workers(self):
        try:
            uid = ResourceActor.default_uid()
            return self.get_actor_ref(uid).get_worker_count()
        except KeyError:
            return 0

    def rescale_workers(self, new_scale, min_workers=None, wait=True, timeout=None):
        min_workers = min_workers or new_scale
        resource_ref = self.get_actor_ref(ResourceActor.default_uid())
        endpoints = self.cluster_info.rescale_workers(new_scale)

        check_start_time = time.time()
        while wait:
            if min_workers <= resource_ref.get_worker_count() <= new_scale:
                break
            self.actor_client.sleep(0.1)
            if timeout and time.time() - check_start_time > timeout:
                raise TimeoutError
        return endpoints

    def get_workers_meta(self):
        resource_uid = ResourceActor.default_uid()
        resource_ref = self.get_actor_ref(resource_uid)
        return resource_ref.get_workers_meta()

    def create_session(self, session_id, **kw):
        self.session_manager.create_session(session_id, **kw)

    def delete_session(self, session_id):
        self.session_manager.delete_session(session_id)

    def has_session(self, session_id):
        """
        Check if the session with given session_id exists.
        """
        return self.session_manager.has_session(session_id)

    def submit_graph(self, session_id, serialized_graph, graph_key, target,
                     names=None, compose=True, wait=True):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        session_ref.submit_tileable_graph(
            serialized_graph, graph_key, target, names=names, compose=compose, _tell=not wait)

    def create_mutable_tensor(self, session_id, name, shape, dtype, *args, **kwargs):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        return session_ref.create_mutable_tensor(name, shape, dtype, *args, **kwargs)

    def get_mutable_tensor(self, session_id, name):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        return session_ref.get_mutable_tensor(name)

    def send_chunk_records(self, session_id, name, chunk_records_to_send, directly=True):
        from .worker.quota import MemQuotaActor
        from .worker.transfer import put_remote_chunk
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)

        chunk_records = []
        for chunk_key, endpoint, records in chunk_records_to_send:
            record_chunk_key = tokenize(chunk_key, uuid.uuid4().hex)
            # register quota
            quota_ref = self.actor_client.actor_ref(MemQuotaActor.default_uid(), address=endpoint)
            quota_ref.request_batch_quota({record_chunk_key: records.nbytes})
            # send record chunk
            receiver_manager_ref = self.actor_client.actor_ref(
                ReceiverManagerActor.default_uid(), address=endpoint)
            put_remote_chunk(session_id, record_chunk_key, records, receiver_manager_ref)
            chunk_records.append((chunk_key, record_chunk_key))

        # register the record chunk to MutableTensorActor
        session_ref.append_chunk_records(name, chunk_records)

    def seal(self, session_id, name):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        return session_ref.seal(name)

    def delete_graph(self, session_id, graph_key):
        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.destroy()

    def stop_graph(self, session_id, graph_key):
        from .scheduler import GraphState
        graph_meta_ref = self.get_graph_meta_ref(session_id, graph_key)
        graph_meta_ref.set_state(GraphState.CANCELLING)

        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.stop_graph()

    def get_graph_state(self, session_id, graph_key):
        from .scheduler import GraphState

        graph_meta_ref = self.get_graph_meta_ref(session_id, graph_key)
        try:
            state_obj = graph_meta_ref.get_state()
            state = state_obj.value if state_obj else 'preparing'
        except ActorNotExist:
            raise GraphNotExists
        return GraphState(state.lower())

    def get_graph_exc_info(self, session_id, graph_key):
        graph_meta_ref = self.get_graph_meta_ref(session_id, graph_key)
        try:
            return graph_meta_ref.get_exc_info()
        except ActorNotExist:
            raise GraphNotExists

    def wait_graph_finish(self, session_id, graph_key, timeout=None):
        graph_meta_ref = self.get_graph_meta_ref(session_id, graph_key)
        self.actor_client.actor_ref(graph_meta_ref.get_wait_ref()).wait(timeout)

    def fetch_chunks_data(self, session_id, chunk_indexes, chunk_keys, nsplits,
                          index_obj=None, serial=True, serial_type=None,
                          compressions=None, pickle_protocol=None):
        chunk_index_to_key = dict((index, key) for index, key in zip(chunk_indexes, chunk_keys))
        if not index_obj:
            chunk_results = dict((idx, self.fetch_chunk_data(session_id, k)) for
                                 idx, k in zip(chunk_indexes, chunk_keys))
        else:
            chunk_results = dict()
            indexes = dict()
            for axis, s in enumerate(index_obj):
                idx_to_slices = slice_split(s, nsplits[axis])
                indexes[axis] = idx_to_slices
            for chunk_index in itertools.product(*[v.keys() for v in indexes.values()]):
                # slice_obj: use tuple, since numpy complains
                #
                # FutureWarning: Using a non-tuple sequence for multidimensional indexing is deprecated; use
                # `arr[tuple(seq)]` instead of `arr[seq]`. In the future this will be interpreted as an array
                # index, `arr[np.array(seq)]`, which will result either in an error or a different result.
                slice_obj = tuple(indexes[axis][chunk_idx] for axis, chunk_idx in enumerate(chunk_index))
                chunk_key = chunk_index_to_key[chunk_index]
                chunk_results[chunk_index] = self.fetch_chunk_data(session_id, chunk_key, slice_obj)

        chunk_results = [(idx, dataserializer.loads(f.result())) for
                         idx, f in chunk_results.items()]
        if len(chunk_results) == 1:
            ret = chunk_results[0][1]
        else:
            ret = merge_chunks(chunk_results)
        if not serial:
            return ret
        compressions = max(compressions) if compressions else dataserializer.CompressType.NONE
        if serial_type == dataserializer.SerialType.PICKLE:
            ret = arrow_array_to_objects(ret)
        return dataserializer.dumps(ret, serial_type=serial_type, compress=compressions,
                                    pickle_protocol=pickle_protocol)

    def fetch_data(self, session_id, tileable_key, index_obj=None, serial=True,
                   serial_type=None, compressions=None, pickle_protocol=None):
        logger.debug('Fetching tileable data %s', tileable_key)
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        graph_ref = self.actor_client.actor_ref(
            session_ref.get_graph_ref_by_tileable_key(tileable_key))
        nsplits, chunk_keys, chunk_indexes = graph_ref.get_tileable_metas(
            [tileable_key], filter_fields=['nsplits', 'chunk_keys', 'chunk_indexes'])[0]
        return self.fetch_chunks_data(session_id, chunk_indexes, chunk_keys, nsplits,
                                      index_obj=index_obj, serial=serial, serial_type=serial_type,
                                      compressions=compressions, pickle_protocol=pickle_protocol)

    def fetch_chunk_data(self, session_id, chunk_key, index_obj=None):
        endpoints = self.chunk_meta_client.get_workers(session_id, chunk_key)
        if endpoints is None:
            raise KeyError(f'Chunk key {chunk_key} not exist in cluster')

        source_endpoint = random.choice(endpoints)
        logger.debug('Fetching chunk %s from worker %s', chunk_key, source_endpoint)
        sender_ref = self.actor_client.actor_ref(ResultSenderActor.default_uid(),
                                                 address=source_endpoint)
        return sender_ref.fetch_data(session_id, chunk_key, index_obj, _wait=False)

    def get_chunk_metas(self, session_id, chunk_keys):
        return self.chunk_meta_client.batch_get_chunk_meta(session_id, chunk_keys)

    def get_tileable_chunk_metas(self, session_id, tileable_key):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        graph_ref = self.actor_client.actor_ref(
            session_ref.get_graph_ref_by_tileable_key(tileable_key))
        return graph_ref.get_tileable_chunk_metas([tileable_key])[0]

    def delete_data(self, session_id, graph_key, tileable_key, wait=False):
        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)
        graph_ref.free_tileable_data(tileable_key, wait=wait, _tell=not wait)

    def get_tileable_nsplits(self, session_id, graph_key, tileable_key):
        # nsplits is essential for operator like `reshape` and shape can be calculated by nsplits
        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)

        return graph_ref.get_tileable_metas([tileable_key], filter_fields=['nsplits'])[0][0]

    def get_tileable_meta(self, session_id, graph_key, tileable_key):
        graph_uid = GraphActor.gen_uid(session_id, graph_key)
        graph_ref = self.get_actor_ref(graph_uid)

        return graph_ref.get_tileable_metas([tileable_key], filter_fields=['nsplits', 'extra_meta'])[0]

    def get_tileable_key_by_name(self, session_id, name):
        session_uid = SessionActor.gen_uid(session_id)
        session_ref = self.get_actor_ref(session_uid)
        return session_ref.get_tileable_key(name)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import os
import re

from .core import FileSystem


magic_check = re.compile('([*?[])')
magic_check_bytes = re.compile(b'([*?[])')


def has_magic(s):
    if isinstance(s, bytes):  # pragma: no cover
        match = magic_check_bytes.search(s)
    else:
        match = magic_check.search(s)
    return match is not None


def _ishidden(path):
    return path[0] in ('.', b'.'[0])


def _isrecursive(pattern):
    if isinstance(pattern, bytes):  # pragma: no cover
        return pattern == b'**'
    else:
        return pattern == '**'


class FileSystemGlob:
    def __init__(self, fs: FileSystem):
        self._fs = fs

    def glob(self, pathname, recursive=False):
        """Return a list of paths matching a pathname pattern.

        The pattern may contain simple shell-style wildcards a la
        fnmatch. However, unlike fnmatch, filenames starting with a
        dot are special cases that are not matched by '*' and '?'
        patterns.

        If recursive is true, the pattern '**' will match any files and
        zero or more directories and subdirectories.
        """
        return list(self.iglob(pathname, recursive=recursive))

    def iglob(self, pathname, recursive=False):
        """Return an iterator which yields the paths matching a pathname pattern.

        The pattern may contain simple shell-style wildcards a la
        fnmatch. However, unlike fnmatch, filenames starting with a
        dot are special cases that are not matched by '*' and '?'
        patterns.

        If recursive is true, the pattern '**' will match any files and
        zero or more directories and subdirectories.
        """
        it = self._iglob(pathname, recursive, False)
        if recursive and _isrecursive(pathname):  # pragma: no cover
            s = next(it)  # skip empty string
            assert not s
        return it

    def _iglob(self, pathname, recursive, dironly):
        dirname, basename = self._fs.path_split(pathname)
        if not has_magic(pathname):
            assert not dironly
            if basename:
                if self._fs.exists(pathname):
                    yield pathname
            else:  # pragma: no cover
                # Patterns ending with a slash should match only directories
                if self._fs.isdir(dirname):
                    yield pathname
            return
        if not dirname:  # pragma: no cover
            if recursive and _isrecursive(basename):
                yield from self._glob2(dirname, basename, dironly)
            else:
                yield from self._glob1(dirname, basename, dironly)
            return
        # `os.path.split()` returns the argument itself as a dirname if it is a
        # drive or UNC path.  Prevent an infinite recursion if a drive or UNC path
        # contains magic characters (i.e. r'\\?\C:').
        if dirname != pathname and has_magic(dirname):
            dirs = self._iglob(dirname, recursive, True)
        else:
            dirs = [dirname]
        if has_magic(basename):
            if recursive and _isrecursive(basename):
                glob_in_dir = self._glob2
            else:
                glob_in_dir = self._glob1
        else:
            glob_in_dir = self._glob0
        for dirname in dirs:
            for name in glob_in_dir(dirname, basename, dironly):
                yield self._fs.path_join(dirname, name)

    # These 2 helper functions non-recursively glob inside a literal directory.
    # They return a list of basenames.  _glob1 accepts a pattern while _glob0
    # takes a literal basename (so it only has to check for its existence).

    def _glob1(self, dirname, pattern, dironly):
        names = list(self._iterdir(dirname, dironly))
        if not _ishidden(pattern):
            names = (x for x in names if not _ishidden(x))
        return fnmatch.filter(names, pattern)

    def _glob0(self, dirname, basename, dironly):  # pragma: no cover
        if not basename:
            # `os.path.split()` returns an empty basename for paths ending with a
            # directory separator.  'q*x/' should match only directories.
            if self._fs.isdir(dirname):
                return [basename]
        else:
            if self._fs.exists(self._fs.path_join(dirname, basename)):
                return [basename]
        return []

    # Following functions are not public but can be used by third-party code.

    def glob0(self, dirname, pattern):  # pragma: no cover
        return self._glob0(dirname, pattern, False)

    def glob1(self, dirname, pattern):  # pragma: no cover
        return self._glob1(dirname, pattern, False)

    # This helper function recursively yields relative pathnames inside a literal
    # directory.

    def _glob2(self, dirname, pattern, dironly):  # pragma: no cover
        assert _isrecursive(pattern)
        yield pattern[:0]
        yield from self._rlistdir(dirname, dironly)

    # If dironly is false, yields all file names inside a directory.
    # If dironly is true, yields only directory names.
    def _iterdir(self, dirname, dironly):
        if not dirname:  # pragma: no cover
            if isinstance(dirname, bytes):
                dirname = bytes(os.curdir, 'ASCII')
            else:
                dirname = os.curdir
        for entry in self._fs.ls(dirname):
            if not dironly or self._fs.isdir(entry):
                yield self._fs.path_split(entry)[-1]

    # Recursively yields relative pathnames inside a literal directory.
    def _rlistdir(self, dirname, dironly):  # pragma: no cover
        names = list(self._iterdir(dirname, dironly))
        for x in names:
            if not _ishidden(x):
                yield x
                path = self._fs.path_join(dirname, x) if dirname else x
                for y in self._rlistdir(path, dironly):
                    yield self._fs.path_join(x, y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorDegrees(TensorUnaryOp):
    _op_type_ = OperandDef.DEGREES
    _func_name = 'degrees'


@infer_dtype(np.degrees)
def degrees(x, out=None, where=None, **kwargs):
    """
    Convert angles from radians to degrees.

    Parameters
    ----------
    x : array_like
        Input tensor in radians.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor of floats
        The corresponding degree values; if `out` was supplied this is a
        reference to it.

    See Also
    --------
    rad2deg : equivalent function

    Examples
    --------
    Convert a radian array to degrees

    >>> import mars.tensor as mt

    >>> rad = mt.arange(12.)*mt.pi/6
    >>> mt.degrees(rad).execute()
    array([   0.,   30.,   60.,   90.,  120.,  150.,  180.,  210.,  240.,
            270.,  300.,  330.])

    >>> out = mt.zeros((rad.shape))
    >>> r = mt.degrees(out)
    >>> mt.all(r == out).execute()
    True
    """
    op = TensorDegrees(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...serialize import BoolField, KeyField
from ..utils import to_numpy
from ..core import TensorOrder
from .core import TensorHasInput


class TensorFromDataFrame(TensorHasInput):
    """ represent tensor from DataFrame """

    _op_type_ = OperandDef.TENSOR_FROM_DATAFRAME
    _input = KeyField('_input')
    _extract_multi_index = BoolField('extract_multi_index')

    def __init__(self, dtype=None, gpu=None, sparse=None, extract_multi_index=False, **kw):
        super().__init__(_dtype=dtype, _gpu=gpu, _sparse=sparse,
                         _extract_multi_index=extract_multi_index, **kw)

    @classmethod
    def execute(cls, ctx, op: 'TensorFromDataFrame'):
        df = ctx[op.inputs[0].key]
        if op._extract_multi_index:
            df = df.to_frame()
        ctx[op.outputs[0].key] = to_numpy(df).astype(op.dtype, order='F')

    @classmethod
    def tile(cls, op: 'TensorFromDataFrame'):
        output = op.outputs[0]

        out_chunks = []
        for c in op.input.chunks:
            shape = (c.shape[0], output.shape[1]) if op._extract_multi_index else c.shape
            index = (c.index[0], 0) if op._extract_multi_index else c.index
            out_chunk = op.copy().reset_key().new_chunk(
                [c], shape=shape, index=index, order=output.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = (op.input.nsplits[0], (output.shape[1],)) if op._extract_multi_index else op.input.nsplits
        return new_op.new_tensors(op.inputs, output.shape, order=output.order,
                                  chunks=out_chunks, nsplits=nsplits)

    def __call__(self, a, order=None):
        from ...dataframe.core import INDEX_TYPE, IndexValue

        if self._extract_multi_index and isinstance(a, INDEX_TYPE) \
                and isinstance(a.index_value.value, IndexValue.MultiIndex):
            order = a.order if order is None else order
            return self.new_tensor([a], (a.shape[0], len(a.index_value.value.names)), order=order)
        else:
            self._extract_multi_index = False

        return super().__call__(a, order=order)


def from_dataframe(in_df, dtype=None):
    from ...dataframe.utils import build_empty_df

    if dtype is None:
        empty_pdf = build_empty_df(in_df.dtypes)
        dtype = to_numpy(empty_pdf).dtype
    op = TensorFromDataFrame(dtype=dtype, gpu=in_df.op.gpu)
    return op(in_df, order=TensorOrder.F_ORDER)  # return tensor with F-order always


def from_series(in_series, dtype=None):
    op = TensorFromDataFrame(dtype=dtype or in_series.dtype, gpu=in_series.op.gpu)
    return op(in_series, order=TensorOrder.F_ORDER)  # return tensor with F-order always


def from_index(in_index, dtype=None, extract_multi_index=False):
    op = TensorFromDataFrame(dtype=dtype or in_index.dtype, gpu=in_index.op.gpu,
                             extract_multi_index=extract_multi_index)
    return op(in_index, order=TensorOrder.F_ORDER)  # return tensor with F-order always



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...serialize import AnyField, BoolField
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, standardize_range_index


class DataFrameExplode(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.EXPLODE

    _column = AnyField('column')
    _ignore_index = BoolField('ignore_field')

    def __init__(self, column=None, ignore_index=None, output_types=None, **kw):
        super().__init__(_column=column, _ignore_index=ignore_index,
                         _output_types=output_types, **kw)

    @property
    def column(self):
        return self._column

    @property
    def ignore_index(self):
        return self._ignore_index

    def _rewrite_params(self, in_obj):
        params = in_obj.params.copy()
        new_shape = list(in_obj.shape)
        new_shape[0] = np.nan
        params['shape'] = tuple(new_shape)

        if self.ignore_index:
            params['index_value'] = parse_index(
                pd.RangeIndex(-1), (in_obj.key, in_obj.index_value.key))
        else:
            params['index_value'] = parse_index(
                None, (in_obj.key, in_obj.index_value.key))
        return params

    def __call__(self, df_or_series):
        return self.new_tileable([df_or_series], **self._rewrite_params(df_or_series))

    @classmethod
    def tile(cls, op: "DataFrameExplode"):
        in_obj = op.inputs[0]

        if in_obj.ndim == 2 and in_obj.chunk_shape[1] > 1:
            # make sure data's second dimension has only 1 chunk
            in_obj = in_obj.rechunk({1: in_obj.shape[1]})._inplace_tile()

        chunks = []
        for chunk in in_obj.chunks:
            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([chunk], **op._rewrite_params(chunk)))

        if op.ignore_index:
            chunks = standardize_range_index(chunks)

        new_op = op.copy().reset_key()
        out_params = op.outputs[0].params
        if in_obj.ndim == 2:
            new_nsplits = ((np.nan,) * in_obj.chunk_shape[0], in_obj.nsplits[1])
        else:
            new_nsplits = ((np.nan,) * in_obj.chunk_shape[0],)
        return new_op.new_tileable([in_obj], chunks=chunks, nsplits=new_nsplits, **out_params)

    @classmethod
    def execute(cls, ctx, op: "DataFrameExplode"):
        in_data = ctx[op.inputs[0].key]
        if in_data.ndim == 2:
            ctx[op.outputs[0].key] = in_data.explode(op.column)
        else:
            ctx[op.outputs[0].key] = in_data.explode()


def df_explode(df, column, ignore_index=False):
    """
    Transform each element of a list-like to a row, replicating index values.

    Parameters
    ----------
    column : str or tuple
        Column to explode.
    ignore_index : bool, default False
        If True, the resulting index will be labeled 0, 1, …, n - 1.

    Returns
    -------
    DataFrame
        Exploded lists to rows of the subset columns;
        index will be duplicated for these rows.

    Raises
    ------
    ValueError :
        if columns of the frame are not unique.

    See Also
    --------
    DataFrame.unstack : Pivot a level of the (necessarily hierarchical)
        index labels.
    DataFrame.melt : Unpivot a DataFrame from wide format to long format.
    Series.explode : Explode a DataFrame from list-like columns to long format.

    Notes
    -----
    This routine will explode list-likes including lists, tuples,
    Series, and np.ndarray. The result dtype of the subset rows will
    be object. Scalars will be returned unchanged. Empty list-likes will
    result in a np.nan for that row.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': [[1, 2, 3], 'foo', [], [3, 4]], 'B': 1})
    >>> df.execute()
               A  B
    0  [1, 2, 3]  1
    1        foo  1
    2         []  1
    3     [3, 4]  1

    >>> df.explode('A').execute()
         A  B
    0    1  1
    0    2  1
    0    3  1
    1  foo  1
    2  NaN  1
    3    3  1
    3    4  1
    """
    op = DataFrameExplode(column=column, ignore_index=ignore_index,
                          output_types=[OutputType.dataframe])
    return op(df)


def series_explode(series, ignore_index=False):
    """
    Transform each element of a list-like to a row.

    Parameters
    ----------
    ignore_index : bool, default False
        If True, the resulting index will be labeled 0, 1, …, n - 1.

    Returns
    -------
    Series
        Exploded lists to rows; index will be duplicated for these rows.

    See Also
    --------
    Series.str.split : Split string values on specified separator.
    Series.unstack : Unstack, a.k.a. pivot, Series with MultiIndex
        to produce DataFrame.
    DataFrame.melt : Unpivot a DataFrame from wide format to long format.
    DataFrame.explode : Explode a DataFrame from list-like
        columns to long format.

    Notes
    -----
    This routine will explode list-likes including lists, tuples,
    Series, and np.ndarray. The result dtype of the subset rows will
    be object. Scalars will be returned unchanged. Empty list-likes will
    result in a np.nan for that row.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> s = md.Series([[1, 2, 3], 'foo', [], [3, 4]])
    >>> s.execute()
    0    [1, 2, 3]
    1          foo
    2           []
    3       [3, 4]
    dtype: object

    >>> s.explode().execute()
    0      1
    0      2
    0      3
    1    foo
    2    NaN
    3      3
    3      4
    dtype: object
    """
    op = DataFrameExplode(ignore_index=ignore_index, output_types=[OutputType.series])
    return op(series)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...config import options
from ...core import OutputType
from ...serialize import BoolField
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameKurtosis(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = opcodes.KURTOSIS
    _func_name = 'kurt'

    _bias = BoolField('bias')
    _fisher = BoolField('fisher')

    def __init__(self, bias=None, fisher=None, **kw):
        super().__init__(_bias=bias, _fisher=fisher, **kw)

    @property
    def bias(self):
        return self._bias

    @property
    def fisher(self):
        return self._fisher

    @classmethod
    def get_reduction_callable(cls, op):
        from .aggregation import where_function
        skipna, bias, fisher = op.skipna, op.bias, op.fisher

        def kurt(x):
            cnt = x.count()
            mean = x.mean(skipna=skipna)
            divided = (x ** 4).mean(skipna=skipna) \
                - 4 * (x ** 3).mean(skipna=skipna) * mean \
                + 6 * (x ** 2).mean(skipna=skipna) * mean ** 2 \
                - 3 * mean ** 4
            var = x.var(skipna=skipna, ddof=0)
            val = where_function(var > 0, divided / var ** 2, np.nan)
            if not bias:
                val = where_function((var > 0) & (cnt > 3),
                                     (val * (cnt ** 2 - 1) - 3 * (cnt - 1) ** 2) / (cnt - 2) / (cnt - 3), np.nan)
            if not fisher:
                val += 3
            return val

        return kurt


def kurt_series(df, axis=None, skipna=None, level=None, combine_size=None, bias=False, fisher=True):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameKurtosis(axis=axis, skipna=skipna, level=level, combine_size=combine_size,
                           bias=bias, fisher=fisher, output_types=[OutputType.scalar],
                           use_inf_as_na=use_inf_as_na)
    return op(df)


def kurt_dataframe(df, axis=None, skipna=None, level=None, numeric_only=None, combine_size=None,
                   bias=False, fisher=True):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameKurtosis(axis=axis, skipna=skipna, level=level, numeric_only=numeric_only,
                           bias=bias, fisher=fisher, combine_size=combine_size,
                           output_types=[OutputType.series], use_inf_as_na=use_inf_as_na)
    return op(df)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...utils import classproperty
from .core import DataFrameUnaryUfunc


class DataFrameNegative(DataFrameUnaryUfunc):
    _op_type_ = OperandDef.NEGATIVE
    _func_name = 'negative'

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorNegative
        return TensorNegative


def negative(df):
    op = DataFrameNegative()
    return op(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import multiprocessing
import os
import signal
import sys
import time

from ...actors import create_actor_pool
from ...cluster_info import StaticClusterDiscoverer
from ...config import options
from ...resource import cpu_count
from ...scheduler.service import SchedulerService
from ...session import new_session
from ...utils import get_next_port, kill_process_tree
from ...worker.service import WorkerService
from .distributor import gen_distributor

try:  # pragma: no cover
    from IPython import get_ipython
    _in_ipython = get_ipython() is not None
except ImportError:
    _in_ipython = False

# forking IPython processes may cause unexpected problem, see issue #1231
_mp_context = multiprocessing.get_context(
    'fork' if not _in_ipython and not sys.platform.startswith('win') else 'spawn')

_local_cluster_clients = dict()
atexit.register(lambda: [v.stop() for v in list(_local_cluster_clients.values())])


class LocalDistributedCluster(object):

    # at least 2 process are required by scheduler and worker
    MIN_SCHEDULER_N_PROCESS = 2
    MIN_WORKER_N_PROCESS = 2

    def __init__(self, endpoint, n_process=None, scheduler_n_process=None,
                 worker_n_process=None, cuda_device=None, ignore_avail_mem=True,
                 shared_memory=None):
        self._endpoint = endpoint

        self._started = False
        self._stopped = False

        self._pool = None
        self._scheduler_service = SchedulerService()
        cuda_devices = [cuda_device] if cuda_device is not None else None
        self._worker_service = WorkerService(ignore_avail_mem=ignore_avail_mem,
                                             cache_mem_size=shared_memory,
                                             cuda_devices=cuda_devices,
                                             distributed=False)

        self._scheduler_n_process, self._worker_n_process = \
            self._calc_scheduler_worker_n_process(n_process,
                                                  scheduler_n_process,
                                                  worker_n_process)

    @property
    def pool(self):
        return self._pool

    @classmethod
    def _calc_scheduler_worker_n_process(cls, n_process, scheduler_n_process, worker_n_process,
                                         calc_cpu_count=cpu_count):
        n_scheduler, n_worker = scheduler_n_process, worker_n_process

        if n_scheduler is None and n_worker is None:
            n_scheduler = cls.MIN_SCHEDULER_N_PROCESS
            n_process = n_process if n_process is not None else calc_cpu_count() + n_scheduler
            n_worker = max(n_process - n_scheduler, cls.MIN_WORKER_N_PROCESS)
        elif n_scheduler is None or n_worker is None:
            # one of scheduler and worker n_process provided
            if n_scheduler is None:
                n_process = n_process if n_process is not None else calc_cpu_count()
                n_scheduler = max(n_process - n_worker, cls.MIN_SCHEDULER_N_PROCESS)
            else:
                assert n_worker is None
                n_process = n_process if n_process is not None else calc_cpu_count() + n_scheduler
                n_worker = max(n_process - n_scheduler, cls.MIN_WORKER_N_PROCESS)

        return n_scheduler, n_worker

    def _make_sure_scheduler_ready(self, timeout=120):
        check_start_time = time.time()
        while True:
            workers_meta = self._scheduler_service._resource_ref.get_workers_meta()
            if not workers_meta:
                # wait for worker to report status
                self._pool.sleep(.5)
                if time.time() - check_start_time > timeout:  # pragma: no cover
                    raise TimeoutError('Check worker ready timed out.')
            else:
                break

    def start_service(self):
        if self._started:
            return
        self._started = True

        # start plasma
        if not options.vineyard.enabled:
            self._worker_service.start_plasma()

        # start actor pool
        n_process = self._scheduler_n_process + self._worker_n_process
        distributor = gen_distributor(self._scheduler_n_process, self._worker_n_process)
        self._pool = create_actor_pool(self._endpoint, n_process, distributor=distributor)

        discoverer = StaticClusterDiscoverer([self._endpoint])

        # start scheduler first
        self._scheduler_service.start(self._endpoint, discoverer, self._pool, distributed=False)

        # start worker next
        self._worker_service.start(self._endpoint, self._pool,
                                   discoverer=discoverer,
                                   process_start_index=self._scheduler_n_process)

        # make sure scheduler is ready
        self._make_sure_scheduler_ready()

    def stop_service(self):
        if self._stopped:
            return

        self._stopped = True
        try:
            self._scheduler_service.stop(self._pool)
            self._worker_service.stop()
        finally:
            self._pool.stop()

    def serve_forever(self):
        try:
            self._pool.join()
        except KeyboardInterrupt:
            pass
        finally:
            self.stop_service()

    def __enter__(self):
        self.start_service()
        return self

    def __exit__(self, *_):
        self.stop_service()


def gen_endpoint(address):
    port = None
    tries = 5  # retry for 5 times

    for i in range(tries):
        try:
            port = get_next_port()
            break
        except SystemError:
            if i < tries - 1:
                continue
            raise

    return f'{address}:{port}'


def _start_cluster(endpoint, event, n_process=None, shared_memory=None, **kw):
    options_dict = kw.pop('options', None) or {}
    options.update(options_dict)

    modules = kw.pop('modules', None) or []
    for m in modules:
        __import__(m, globals(), locals(), [])

    cluster = LocalDistributedCluster(endpoint, n_process=n_process,
                                      shared_memory=shared_memory, **kw)
    cluster.start_service()
    event.set()
    try:
        cluster.serve_forever()
    finally:
        cluster.stop_service()


def _start_cluster_process(endpoint, n_process, shared_memory, **kw):
    event = _mp_context.Event()

    kw = kw.copy()
    kw['n_process'] = n_process
    kw['shared_memory'] = shared_memory or '20%'
    process = _mp_context.Process(
        target=_start_cluster, args=(endpoint, event), kwargs=kw)
    process.start()

    while True:
        event.wait(5)
        if not event.is_set():
            # service not started yet
            continue
        if not process.is_alive():
            raise SystemError('New local cluster failed')
        else:
            break

    return process


def _start_web(scheduler_address, ui_port, event, options_dict):
    import gevent.monkey
    gevent.monkey.patch_all(thread=False)

    options.update(options_dict)

    from ...web import MarsWeb

    web = MarsWeb(None, ui_port, scheduler_address)
    try:
        web.start(event=event, block=True)
    finally:
        web.stop()


def _start_web_process(scheduler_endpoint, web_endpoint):
    ui_port = int(web_endpoint.rsplit(':', 1)[1])
    options_dict = options.to_dict()

    web_event = _mp_context.Event()
    web_process = _mp_context.Process(
        target=_start_web, args=(scheduler_endpoint, ui_port, web_event, options_dict),
        daemon=True)
    web_process.start()

    while True:
        web_event.wait(5)
        if not web_event.is_set():
            # web not started yet
            continue
        if not web_process.is_alive():
            raise SystemError('New web interface failed')
        else:
            break

    return web_process


class LocalDistributedClusterClient(object):
    def __init__(self, endpoint, web_endpoint, cluster_process, web_process):
        self._cluster_process = cluster_process
        self._web_process = web_process
        self._endpoint = endpoint
        self._web_endpoint = web_endpoint
        self._session = new_session(endpoint).as_default()

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def web_endpoint(self):
        return self._web_endpoint

    @property
    def session(self):
        return self._session

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    @staticmethod
    def _ensure_process_finish(proc):
        if proc is None or not proc.is_alive():
            return
        proc.join(3)
        kill_process_tree(proc.pid)

    def stop(self):
        try:
            del _local_cluster_clients[id(self)]
        except KeyError:  # pragma: no cover
            pass

        if self._cluster_process.is_alive():
            os.kill(self._cluster_process.pid, signal.SIGINT)
        if self._web_process is not None and self._web_process.is_alive():
            os.kill(self._web_process.pid, signal.SIGINT)

        self._ensure_process_finish(self._cluster_process)
        self._ensure_process_finish(self._web_process)


def new_cluster(address='0.0.0.0', web=False, n_process=None, shared_memory=None,
                open_browser=None, **kw):
    open_browser = open_browser if open_browser is not None else options.deploy.open_browser
    endpoint = gen_endpoint(address)
    web_endpoint = None
    if web is True:
        web_endpoint = gen_endpoint('0.0.0.0')
    elif isinstance(web, str):
        if ':' in web:
            web_endpoint = web
        else:
            web_endpoint = gen_endpoint(web)

    options_dict = options.to_dict()
    options_dict.update(kw.get('options') or dict())
    kw['options'] = options_dict

    process = _start_cluster_process(endpoint, n_process, shared_memory, **kw)

    web_process = None
    if web_endpoint:
        web_process = _start_web_process(endpoint, web_endpoint)
        print(f'Web endpoint started at http://{web_endpoint}', file=sys.stderr)
        if open_browser:
            import webbrowser
            webbrowser.open_new_tab(f'http://{web_endpoint}')

    client = LocalDistributedClusterClient(endpoint, web_endpoint, process, web_process)
    _local_cluster_clients[id(client)] = client
    return client



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..config import options
from ..utils import log_unhandled
from .storage import DataStorageDevice
from .utils import WorkerActor


class SealActor(WorkerActor):
    """
    Actor sealing a chunk from a serials of record chunks.
    """
    @staticmethod
    def gen_uid(session_id, chunk_key):
        return f's:0:seal${session_id}${chunk_key}'

    def __init__(self):
        super().__init__()
        self._mem_quota_ref = None

    def post_create(self):
        from .quota import MemQuotaActor
        super().post_create()
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_uid())

    @log_unhandled
    def seal_chunk(self, session_id, graph_key, chunk_key, keys, shape, record_type, dtype, fill_value):
        from ..scheduler.chunkmeta import WorkerMeta

        chunk_bytes_size = np.prod(shape) * dtype.itemsize
        self._mem_quota_ref.request_batch_quota({chunk_key: chunk_bytes_size})
        if fill_value is None:
            ndarr = np.zeros(shape, dtype=dtype)
        else:
            ndarr = np.full(shape, fill_value, dtype=dtype)
        ndarr_ts = np.zeros(shape, dtype=np.dtype('datetime64[ns]'))

        if options.vineyard.enabled:
            in_memory_device = DataStorageDevice.VINEYARD
        else:
            in_memory_device = DataStorageDevice.SHARED_MEMORY
        use_devices = [in_memory_device, DataStorageDevice.DISK]  # pragma: no cover

        # consolidate
        for key in keys:
            buffer = None
            try:
                # todo potential memory quota issue must be dealt with
                obj = self.storage_client.get_object(
                    session_id, key, use_devices, _promise=False)
                record_view = obj.view(dtype=record_type, type=np.recarray)

                for record in record_view:
                    idx = np.unravel_index(record.index, shape)
                    if record.ts > ndarr_ts[idx]:
                        ndarr[idx] = record.value
            finally:
                del buffer

            # clean up
            self.storage_client.delete(session_id, [key])
            self.get_meta_client().delete_meta(session_id, key, False)

        self._mem_quota_ref.release_quotas(keys)

        self.storage_client.put_objects(
            session_id, [chunk_key], [ndarr], use_devices)
        worker_meta = WorkerMeta(chunk_bytes_size, shape, (self.address,))
        self.get_meta_client().batch_set_chunk_meta(session_id, [chunk_key], [worker_meta], _wait=False)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from ... import opcodes
from ...core import get_output_types, OutputType
from ...serialize import AnyField, StringField
from ..core import SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_df, build_series, validate_axis, parse_index


class DataFrameRename(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.RENAME

    _columns_mapper = AnyField('columns_mapper')
    _index_mapper = AnyField('index_mapper')
    _new_name = AnyField('new_name')
    _level = AnyField('level')
    _errors = StringField('errors')

    def __init__(self, columns_mapper=None, index_mapper=None, new_name=None, level=None,
                 errors=None, output_types=None, **kw):
        super().__init__(_columns_mapper=columns_mapper, _index_mapper=index_mapper,
                         _new_name=new_name, _level=level, _errors=errors,
                         _output_types=output_types, **kw)

    @property
    def columns_mapper(self):
        return self._columns_mapper

    @property
    def index_mapper(self):
        return self._index_mapper

    @property
    def new_name(self):
        return self._new_name

    @property
    def level(self):
        return self._level

    @property
    def errors(self) -> str:
        return self._errors

    def _calc_renamed_df(self, df, errors='ignore'):
        empty_df = build_df(df)
        return empty_df.rename(columns=self._columns_mapper, index=self._index_mapper,
                               level=self._level, errors=errors)

    def _calc_renamed_series(self, df, errors='ignore'):
        empty_series = build_series(df, name=df.name)
        new_series = empty_series.rename(index=self._index_mapper, level=self._level, errors=errors)
        if self._new_name:
            new_series.name = self._new_name
        return new_series

    def __call__(self, df):
        params = df.params
        raw_index = df.index_value.to_pandas()
        if df.ndim == 2:
            new_df = self._calc_renamed_df(df, errors=self.errors)
            new_index = new_df.index
        elif isinstance(df, SERIES_TYPE):
            new_df = self._calc_renamed_series(df, errors=self.errors)
            new_index = new_df.index
        else:
            new_df = new_index = raw_index.set_names(self._index_mapper or self._new_name,
                                                     level=self._level)

        if self._columns_mapper is not None:
            params['columns_value'] = parse_index(new_df.columns, store_data=True)
            params['dtypes'] = new_df.dtypes
        if self._index_mapper is not None:
            params['index_value'] = parse_index(new_index)
        if df.ndim == 1:
            params['name'] = new_df.name
        return self.new_tileable([df], **params)

    @classmethod
    def tile(cls, op: 'DataFrameRename'):
        inp = op.inputs[0]
        out = op.outputs[0]
        chunks = []

        dtypes_cache = dict()
        for c in inp.chunks:
            params = c.params
            new_op = op.copy().reset_key()

            if op.columns_mapper is not None:
                try:
                    new_dtypes = dtypes_cache[c.index[0]]
                except KeyError:
                    new_dtypes = dtypes_cache[c.index[0]] = op._calc_renamed_df(c).dtypes

                params['columns_value'] = parse_index(new_dtypes.index, store_data=True)
                params['dtypes'] = new_dtypes
            if op.index_mapper is not None:
                params['index_value'] = out.index_value
            if op.new_name is not None:
                params['name'] = out.name

            if isinstance(op.columns_mapper, dict):
                idx = params['dtypes'].index
                if op._level is not None:
                    idx = idx.get_level_values(op._level)
                new_op._columns_mapper = {k: v for k, v in op.columns_mapper.items()
                                          if v in idx}
            chunks.append(new_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        return new_op.new_tileables([inp], chunks=chunks, nsplits=inp.nsplits, **out.params)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameRename'):
        input_ = ctx[op.inputs[0].key]
        if input_.ndim == 2:
            ctx[op.outputs[0].key] = input_.rename(index=op.index_mapper, columns=op.columns_mapper,
                                                   level=op.level)
        elif op.output_types[0] == OutputType.series:
            ctx[op.outputs[0].key] = input_.rename(index=op.index_mapper or op.new_name, level=op.level)
        else:
            ctx[op.outputs[0].key] = input_.set_names(op.index_mapper or op.new_name,
                                                      level=op.level)


def _rename(df_obj, index_mapper=None, columns_mapper=None, copy=True, inplace=False,
            level=None, errors='ignore'):
    if not copy:
        raise NotImplementedError('`copy=False` not implemented')

    if index_mapper is not None and errors == 'raise' and not inplace:
        warnings.warn('Errors will not raise for non-existing indices')

    op = DataFrameRename(columns_mapper=columns_mapper, index_mapper=index_mapper,
                         level=level, errors=errors, output_types=get_output_types(df_obj))
    ret = op(df_obj)
    if inplace:
        df_obj.data = ret.data
    else:
        return ret


def df_rename(df, mapper=None, index=None, columns=None, axis='index', copy=True,
              inplace=False, level=None, errors='ignore'):
    """
    Alter axes labels.

    Function / dict values must be unique (1-to-1). Labels not contained in
    a dict / Series will be left as-is. Extra labels listed don't throw an
    error.

    Parameters
    ----------
    mapper : dict-like or function
        Dict-like or functions transformations to apply to
        that axis' values. Use either ``mapper`` and ``axis`` to
        specify the axis to target with ``mapper``, or ``index`` and
        ``columns``.
    index : dict-like or function
        Alternative to specifying axis (``mapper, axis=0``
        is equivalent to ``index=mapper``).
    columns : dict-like or function
        Alternative to specifying axis (``mapper, axis=1``
        is equivalent to ``columns=mapper``).
    axis : int or str
        Axis to target with ``mapper``. Can be either the axis name
        ('index', 'columns') or number (0, 1). The default is 'index'.
    copy : bool, default True
        Also copy underlying data.
    inplace : bool, default False
        Whether to return a new DataFrame. If True then value of copy is
        ignored.
    level : int or level name, default None
        In case of a MultiIndex, only rename labels in the specified
        level.
    errors : {'ignore', 'raise'}, default 'ignore'
        If 'raise', raise a `KeyError` when a dict-like `mapper`, `index`,
        or `columns` contains labels that are not present in the Index
        being transformed.
        If 'ignore', existing keys will be renamed and extra keys will be
        ignored.

    Returns
    -------
    DataFrame
        DataFrame with the renamed axis labels.

    Raises
    ------
    KeyError
        If any of the labels is not found in the selected axis and
        "errors='raise'".

    See Also
    --------
    DataFrame.rename_axis : Set the name of the axis.

    Examples
    --------

    ``DataFrame.rename`` supports two calling conventions

    * ``(index=index_mapper, columns=columns_mapper, ...)``
    * ``(mapper, axis={'index', 'columns'}, ...)``

    We *highly* recommend using keyword arguments to clarify your
    intent.

    Rename columns using a mapping:

    >>> import mars.dataframe as md
    >>> df = md.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    >>> df.rename(columns={"A": "a", "B": "c"}).execute()
       a  c
    0  1  4
    1  2  5
    2  3  6

    Rename index using a mapping:

    >>> df.rename(index={0: "x", 1: "y", 2: "z"}).execute()
       A  B
    x  1  4
    y  2  5
    z  3  6

    Cast index labels to a different type:

    >>> df.index.execute()
    RangeIndex(start=0, stop=3, step=1)
    >>> df.rename(index=str).index.execute()
    Index(['0', '1', '2'], dtype='object')

    >>> df.rename(columns={"A": "a", "B": "b", "C": "c"}, errors="raise").execute()
    Traceback (most recent call last):
    KeyError: ['C'] not found in axis

    Using axis-style parameters

    >>> df.rename(str.lower, axis='columns').execute()
       a  b
    0  1  4
    1  2  5
    2  3  6

    >>> df.rename({1: 2, 2: 4}, axis='index').execute()
       A  B
    0  1  4
    2  2  5
    4  3  6

    """
    axis = validate_axis(axis, df)
    if axis == 0:
        index_mapper = index if index is not None else mapper
        columns_mapper = columns
    else:
        columns_mapper = columns if columns is not None else mapper
        index_mapper = index

    if index_mapper is not None and errors == 'raise' and not inplace:
        warnings.warn('Errors will not raise for non-existing indices')

    return _rename(df, index_mapper=index_mapper, columns_mapper=columns_mapper, copy=copy,
                   inplace=inplace, level=level, errors=errors)


def series_rename(series, index=None, *, axis='index', copy=True, inplace=False, level=None,
                  errors='ignore'):
    """
    Alter Series index labels or name.

    Function / dict values must be unique (1-to-1). Labels not contained in
    a dict / Series will be left as-is. Extra labels listed don't throw an
    error.

    Alternatively, change ``Series.name`` with a scalar value.

    Parameters
    ----------
    axis : {0 or "index"}
        Unused. Accepted for compatability with DataFrame method only.
    index : scalar, hashable sequence, dict-like or function, optional
        Functions or dict-like are transformations to apply to
        the index.
        Scalar or hashable sequence-like will alter the ``Series.name``
        attribute.

    **kwargs
        Additional keyword arguments passed to the function. Only the
        "inplace" keyword is used.

    Returns
    -------
    Series
        Series with index labels or name altered.

    See Also
    --------
    DataFrame.rename : Corresponding DataFrame method.
    Series.rename_axis : Set the name of the axis.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 2, 3])
    >>> s.execute()
    0    1
    1    2
    2    3
    dtype: int64
    >>> s.rename("my_name").execute()  # scalar, changes Series.name.execute()
    0    1
    1    2
    2    3
    Name: my_name, dtype: int64
    >>> s.rename(lambda x: x ** 2).execute()  # function, changes labels.execute()
    0    1
    1    2
    4    3
    dtype: int64
    >>> s.rename({1: 3, 2: 5}).execute()  # mapping, changes labels.execute()
    0    1
    3    2
    5    3
    dtype: int64
    """
    validate_axis(axis)
    return _rename(series, index_mapper=index, copy=copy, inplace=inplace, level=level,
                   errors=errors)


def index_rename(index, name, inplace=False):
    """
    Alter Index or MultiIndex name.

    Able to set new names without level. Defaults to returning new index.
    Length of names must match number of levels in MultiIndex.

    Parameters
    ----------
    name : label or list of labels
        Name(s) to set.
    inplace : bool, default False
        Modifies the object directly, instead of creating a new Index or
        MultiIndex.

    Returns
    -------
    Index
        The same type as the caller or None if inplace is True.

    See Also
    --------
    Index.set_names : Able to set new names partially and by level.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> idx = md.Index(['A', 'C', 'A', 'B'], name='score')
    >>> idx.rename('grade').execute()
    Index(['A', 'C', 'A', 'B'], dtype='object', name='grade')

    >>> idx = md.Index([('python', 2018),
    ...                 ('python', 2019),
    ...                 ('cobra', 2018),
    ...                 ('cobra', 2019)],
    ...                names=['kind', 'year'])
    >>> idx.execute()
    MultiIndex([('python', 2018),
                ('python', 2019),
                ( 'cobra', 2018),
                ( 'cobra', 2019)],
               names=['kind', 'year'])
    >>> idx.rename(['species', 'year']).execute()
    MultiIndex([('python', 2018),
                ('python', 2019),
                ( 'cobra', 2018),
                ( 'cobra', 2019)],
               names=['species', 'year'])
    >>> idx.rename('species').execute()
    Traceback (most recent call last):
    TypeError: Must pass list-like as `names`.
    """
    op = DataFrameRename(index_mapper=name, output_types=get_output_types(index))
    ret = op(index)
    if inplace:
        index.data = ret.data
    else:
        return ret


def index_set_names(index, names, level=None, inplace=False):
    """
    Set Index or MultiIndex name.

    Able to set new names partially and by level.

    Parameters
    ----------
    names : label or list of label
        Name(s) to set.
    level : int, label or list of int or label, optional
        If the index is a MultiIndex, level(s) to set (None for all
        levels). Otherwise level must be None.
    inplace : bool, default False
        Modifies the object directly, instead of creating a new Index or
        MultiIndex.

    Returns
    -------
    Index
        The same type as the caller or None if inplace is True.

    See Also
    --------
    Index.rename : Able to set new names without level.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> idx = md.Index([1, 2, 3, 4])
    >>> idx.execute()
    Int64Index([1, 2, 3, 4], dtype='int64')
    >>> idx.set_names('quarter').execute()
    Int64Index([1, 2, 3, 4], dtype='int64', name='quarter')

    >>> idx = md.MultiIndex.from_product([['python', 'cobra'],
    ...                                   [2018, 2019]])
    >>> idx.execute()
    MultiIndex([('python', 2018),
                ('python', 2019),
                ( 'cobra', 2018),
                ( 'cobra', 2019)],
               )
    >>> idx.set_names(['kind', 'year'], inplace=True)
    >>> idx.execute()
    MultiIndex([('python', 2018),
                ('python', 2019),
                ( 'cobra', 2018),
                ( 'cobra', 2019)],
               names=['kind', 'year'])
    >>> idx.set_names('species', level=0).execute()
    MultiIndex([('python', 2018),
                ('python', 2019),
                ( 'cobra', 2018),
                ( 'cobra', 2019)],
               names=['species', 'year'])
    """
    op = DataFrameRename(index_mapper=names, level=level,
                         output_types=get_output_types(index))
    ret = op(index)

    if inplace:
        df_or_series = getattr(index, '_get_df_or_series', lambda: None)()
        if df_or_series is not None:
            from .rename_axis import rename_axis_with_level
            rename_axis_with_level(df_or_series, names, axis=index._axis,
                                   level=level, inplace=True)
            index.data = df_or_series.axes[index._axis].data
        else:
            index.data = ret.data
    else:
        return ret



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import random
import time
from collections import defaultdict, deque
from typing import NamedTuple, Optional

from ...config import options
from ...errors import SpillNotConfigured
from ...utils import tokenize, mod_hash
from ..utils import WorkerActor, parse_spill_dirs

logger = logging.getLogger(__name__)


def _build_file_path_by_name(session_id, filename):
    dirs = options.worker.spill_directory
    dir_name = dirs[mod_hash(filename, len(dirs))]
    return os.path.join(dir_name, str(session_id), filename)


class DataMeta(NamedTuple):
    filename: str
    start: int
    end: Optional[int]


class DiskFileMergerActor(WorkerActor):
    def __init__(self):
        super().__init__()

        dirs = options.worker.spill_directory = parse_spill_dirs(options.worker.spill_directory)
        if not dirs:  # pragma: no cover
            raise SpillNotConfigured

        self._key_to_data_meta = dict()
        self._file_to_sizes = defaultdict(lambda: 0)
        self._file_to_keys = defaultdict(list)
        self._available_files = []
        self._concurrency = options.worker.filemerger.concurrency
        self._max_file_size = options.worker.filemerger.max_file_size

        self._reading_files = set()
        self._writing_files = set()

        self._pending_read_requests = deque()
        self._pending_write_requests = deque()

    def await_file_reader(self, session_id, data_key, with_lock=False, callback=None):
        if with_lock and len(self._reading_files) >= self._concurrency:
            self._pending_read_requests.append((session_id, data_key, callback))
            return
        self._reading_files.add((session_id, data_key))
        self.tell_promise(callback)

    def await_file_writer(self, session_id, with_lock=True, callback=None):
        if with_lock and len(self._writing_files) >= self._concurrency:
            self._pending_write_requests.append(callback)
            return

        if self._available_files:
            filename = self._available_files.pop(-1)
        else:
            filename = _build_file_path_by_name(
                session_id, tokenize(callback, time.time(), random.random()))

        os.makedirs(os.path.dirname(filename), exist_ok=True)
        self._writing_files.add(filename)
        self.tell_promise(callback, filename, self._file_to_sizes[filename])

    def release_file_reader(self, session_id, data_key):
        self._reading_files.difference_update([(session_id, data_key)])
        if self._pending_read_requests:
            session_id, data_key, callback = self._pending_read_requests.popleft()
            self.await_file_reader(session_id, data_key, callback)

    def release_file_writer(self, session_id, data_key, filename, start, end):
        self._writing_files.remove(filename)
        self._key_to_data_meta[(session_id, data_key)] = DataMeta(filename, start, end)
        self._file_to_sizes[filename] = end
        self._file_to_keys[filename].append((session_id, data_key))

        if self._file_to_sizes[filename] >= self._max_file_size:
            self._file_to_sizes.pop(filename)
        else:
            self._available_files.append(filename)

        if self._pending_write_requests:
            self.await_file_writer(session_id, callback=self._pending_write_requests.popleft())

    def get_file_metas(self, session_id, data_keys):
        metas = []
        for key in data_keys:
            meta = self._key_to_data_meta.get((session_id, key))
            metas.append(meta)
        return metas

    def delete_file_metas(self, session_id, data_keys):
        files_to_delete = []
        filtered_keys = []
        for key in data_keys:
            meta = self._key_to_data_meta.pop((session_id, key), None)
            if meta is None:  # pragma: no cover
                continue
            filtered_keys.append(key)
            filename = meta.filename
            keys_set = self._file_to_keys[filename]
            keys_set.remove((session_id, key))
            if not keys_set:
                if filename in self._writing_files:
                    continue
                del self._file_to_keys[filename]
                self._file_to_sizes.pop(filename, None)
                files_to_delete.append(filename)

        del_set = set(files_to_delete)
        if del_set:
            self._available_files = [f for f in self._available_files if f not in del_set]
        return files_to_delete, filtered_keys

    def dump_info(self):
        return self._key_to_data_meta, self._file_to_sizes, self._file_to_keys, \
            self._writing_files



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...serialize import DataFrameField, SeriesField
from ...tensor.utils import get_chunk_slices
from ..utils import decide_dataframe_chunk_sizes, parse_index
from ..operands import DataFrameOperand, DataFrameOperandMixin


class DataFrameDataSource(DataFrameOperand, DataFrameOperandMixin):
    """
    Represents data from pandas DataFrame
    """

    _op_type_ = OperandDef.DATAFRAME_DATA_SOURCE

    _data = DataFrameField('data')
    _dtypes = SeriesField('dtypes')

    def __init__(self, data=None, dtypes=None, gpu=None, sparse=None, **kw):
        if dtypes is None and data is not None:
            dtypes = data.dtypes
        super().__init__(_data=data, _dtypes=dtypes, _gpu=gpu, _sparse=sparse,
                         _output_types=[OutputType.dataframe], **kw)

    @property
    def data(self):
        return self._data

    @property
    def dtypes(self):
        return self._dtypes

    def __call__(self, shape, chunk_size=None):
        return self.new_dataframe(None, shape, dtypes=self.dtypes,
                                  index_value=parse_index(self._data.index),
                                  columns_value=parse_index(self._data.columns,
                                                            store_data=True),
                                  raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        df = op.outputs[0]
        raw_df = op.data

        memory_usage = raw_df.memory_usage(index=False, deep=True)
        chunk_size = df.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_dataframe_chunk_sizes(df.shape, chunk_size, memory_usage)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        index_values = dict()
        column_values = dict()
        for chunk_shape, chunk_idx in zip(itertools.product(*chunk_size),
                                          itertools.product(*chunk_size_idxes)):
            chunk_op = op.copy().reset_key()
            slc = get_chunk_slices(chunk_size, chunk_idx)
            i_slc, j_slc = slc
            if j_slc == slice(0, df.shape[1]):
                # optimize full slice, it's way more faster
                j_slc = slice(None)
            chunk_op._data = raw_df.iloc[i_slc, j_slc]
            chunk_op._dtypes = chunk_op._data.dtypes
            i, j = chunk_idx
            if i in index_values:
                index_value = index_values[i]
            else:
                index_value = index_values[i] = parse_index(chunk_op.data.index)
            if j in column_values:
                column_value = column_values[j]
            else:
                column_value = column_values[j] = parse_index(chunk_op.data.columns,
                                                              store_data=True)
            out_chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx,
                                           index_value=index_value,
                                           columns_value=column_value,
                                           dtypes=chunk_op._data.dtypes)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(None, df.shape, dtypes=op.dtypes,
                                     index_value=df.index_value,
                                     columns_value=df.columns_value,
                                     chunks=out_chunks, nsplits=chunk_size)

    @classmethod
    def execute(cls, ctx, op):
        ctx[op.outputs[0].key] = op.data


def from_pandas(data, chunk_size=None, gpu=False, sparse=False):
    op = DataFrameDataSource(data=data, gpu=gpu, sparse=sparse)
    return op(data.shape, chunk_size=chunk_size)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from .... import opcodes as OperandDef
from ....operands import OperandStage
from ....serialize import ValueType, KeyField, BoolField, TupleField
from ....tiles import TilesError
from ....config import options
from ....utils import check_chunks_unknown_shape, get_shuffle_input_keys_idxes, \
    require_module, recursive_tile
from ...core import TensorOrder
from ...operands import TensorMapReduceOperand, TensorOperandMixin, TensorShuffleProxy
from ...datasource import ascontiguousarray, array, zeros
from ...arithmetic import equal
from ...utils import decide_chunk_sizes
from ...array_utils import as_same_device, device, cp


class TensorSquareform(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.SQUAREFORM

    _input = KeyField('input')
    _checks = BoolField('checks')

    _checks_input = KeyField('checks_input')
    _x_shape = TupleField('x_shape', ValueType.int32)
    _reduce_sizes = TupleField('reduce_sizes', ValueType.tuple)
    _start_positions = TupleField('start_positions', ValueType.int32)

    def __init__(self, stage=None, checks=None, checks_input=None, x_shape=None,
                 reduce_sizes=None, start_positions=None, shuffle_key=None,
                 dtype=None, gpu=None, **kw):
        super().__init__(_stage=stage, _checks=checks, _checks_input=checks_input,
                         _x_shape=x_shape, _reduce_sizes=reduce_sizes,
                         _start_positions=start_positions, _shuffle_key=shuffle_key,
                         _dtype=dtype, _gpu=gpu, **kw)

    @property
    def input(self):
        return self._input

    @property
    def checks(self):
        return self._checks

    @property
    def checks_input(self):
        return self._checks_input

    @property
    def x_shape(self):
        return self._x_shape

    @property
    def reduce_sizes(self):
        return self._reduce_sizes

    @property
    def start_positions(self):
        return self._start_positions

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if self._checks_input is not None:
            self._checks_input = self._inputs[-1]

    def __call__(self, X, force='no', chunk_size=None):
        s = X.shape

        if force.lower() == 'tomatrix':
            if len(s) != 1:
                raise ValueError("Forcing 'tomatrix' but input X is not a "
                                 "distance vector.")
        elif force.lower() == 'tovector':
            if len(s) != 2:
                raise ValueError("Forcing 'tovector' but input X is not a "
                                 "distance matrix.")

        # X = squareform(v)
        if len(s) == 1:
            if s[0] == 0:
                return zeros((1, 1), dtype=X.dtype)

            # Grab the closest value to the square root of the number
            # of elements times 2 to see if the number of elements
            # is indeed a binomial coefficient.
            d = int(np.ceil(np.sqrt(s[0] * 2)))

            # Check that v is of valid dimensions.
            if d * (d - 1) != s[0] * 2:
                raise ValueError('Incompatible vector size. It must be a binomial '
                                 'coefficient n choose 2 for some integer n >= 2.')

            shape = (d, d)
        elif len(s) == 2:
            if s[0] != s[1]:
                raise ValueError('The matrix argument must be square.')

            # One-side of the dimensions is set here.
            d = s[0]

            if d <= 1:
                return array([], dtype=X.dtype)

            shape = ((d * (d - 1)) // 2,)
        else:
            raise ValueError(('The first argument must be one or two dimensional '
                              'tensor. A %d-dimensional tensor is not '
                              'permitted') % len(s))

        return self.new_tensor([X], shape=shape, order=TensorOrder.C_ORDER,
                               raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_chunk_sizes(tensor.shape, chunk_size, tensor.dtype.itemsize)
        n_chunk = np.product([len(cs) for cs in chunk_size])

        if len(op.input.chunks) == 1 and n_chunk == 1:
            return cls._tile_one_chunk(op)
        else:
            return cls._tile_chunks(op, chunk_size)

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk = chunk_op.new_chunk(op.input.chunks, shape=out.shape,
                                   order=out.order,
                                   index=(0,) * out.ndim)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape,
                                  order=out.order, chunks=[chunk],
                                  nsplits=tuple((s,) for s in out.shape))

    @classmethod
    def _gen_checks_input(cls, op):
        if op.input.ndim != 2 or not op.checks:
            return

        x = op.input
        return recursive_tile(equal(x, x.T).all()).chunks[0]

    @classmethod
    def _tile_chunks(cls, op, chunk_size):
        check_chunks_unknown_shape(op.inputs, TilesError)
        out = op.outputs[0]

        checks_input = cls._gen_checks_input(op)

        map_chunks = []
        cum_sizes = [[0] + np.cumsum(ns).tolist()
                     for ns in op.input.nsplits]
        to_vec = op.input.ndim == 2
        for in_chunk in op.input.chunks:
            if to_vec and in_chunk.index[0] > in_chunk.index[1]:
                # if apply squareform to 2-d tensor which is symmetric,
                # we don't need to calculate for lower triangle chunks
                continue
            map_chunk_op = TensorSquareform(
                stage=OperandStage.map, checks_input=checks_input, reduce_sizes=chunk_size,
                x_shape=op.input.shape,
                start_positions=tuple(cum_sizes[ax][j]
                                      for ax, j in enumerate(in_chunk.index)),
                dtype=out.dtype, gpu=out.op.gpu)
            chunk_inputs = [in_chunk]
            if checks_input is not None:
                chunk_inputs.append(checks_input)
            map_chunk = map_chunk_op.new_chunk(chunk_inputs, shape=(2, np.nan),
                                               index=in_chunk.index,
                                               order=out.order)
            map_chunks.append(map_chunk)

        proxy_chunk = TensorShuffleProxy(dtype=out.dtype).new_chunk(
            map_chunks, shape=())

        reduce_chunks = []
        out_shape_iter = itertools.product(*chunk_size)
        out_idx_iter = itertools.product(*(range(len(cs)) for cs in chunk_size))
        for out_idx, out_shape in zip(out_idx_iter, out_shape_iter):
            reduce_chunk_op = TensorSquareform(
                stage=OperandStage.reduce,
                shuffle_key=','.join(str(i) for i in out_idx),
                dtype=out.dtype)
            reduce_chunk = reduce_chunk_op.new_chunk(
                [proxy_chunk], shape=out_shape, index=out_idx, order=out.order)
            reduce_chunks.append(reduce_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  nsplits=chunk_size, chunks=reduce_chunks)

    @classmethod
    def _to_matrix(cls, ctx, xp, x, op):
        assert x.ndim == 1
        out_chunk_size = op.reduce_sizes
        out_shape = tuple(sum(ns) for ns in out_chunk_size)
        d = out_shape[0]

        # calculate the index for the 1-d chunk
        index = xp.arange(x.shape[0])
        index = xp.add(index, op.start_positions[0], out=index)

        # input length for each row
        row_sizes = xp.arange(d, -1, -1)
        row_sizes[0] = 0
        xp.cumsum(row_sizes[1:], out=row_sizes[1:])
        # calculate row for each element
        rows = xp.searchsorted(row_sizes, index, side='right')
        xp.subtract(rows, 1, out=rows)
        # calculate col for each element
        # offsets
        cols_offsets = xp.arange(1, d + 1)
        cols = xp.empty(x.shape, dtype=np.int32)
        xp.add(xp.subtract(index, row_sizes[rows], out=cols),
               cols_offsets[rows], out=cols)

        cum_sizes = [[0] + np.cumsum(cs).tolist() for cs in out_chunk_size]
        for idx in itertools.product(*(range(len(ns)) for ns in out_chunk_size)):
            shuffle_key = ','.join(str(p) for p in idx)
            i, j = idx
            row_range = cum_sizes[0][i], cum_sizes[0][i + 1]
            col_range = cum_sizes[1][j], cum_sizes[1][j + 1]
            # for upper
            filtered = (rows >= row_range[0]) & (rows < row_range[1]) & \
                       (cols >= col_range[0]) & (cols < col_range[1])
            inds_tup = rows[filtered] - row_range[0], cols[filtered] - col_range[0]
            upper_inds = xp.ravel_multi_index(inds_tup, (out_chunk_size[0][i], out_chunk_size[1][j]))
            upper_values = x[filtered]
            # for lower
            filtered = (rows >= col_range[0]) & (rows < col_range[1]) & \
                       (cols >= row_range[0]) & (cols < row_range[1])
            inds_tup = cols[filtered] - row_range[0], rows[filtered] - col_range[0]
            lower_inds = xp.ravel_multi_index(inds_tup, (out_chunk_size[0][i], out_chunk_size[1][j]))
            lower_values = x[filtered]

            inds = xp.concatenate([upper_inds, lower_inds])
            values = xp.concatenate([upper_values, lower_values])

            ctx[op.outputs[0].key, shuffle_key] = inds, values

    @classmethod
    def _to_vector(cls, ctx, xp, x, op):
        out_chunk_size = op.reduce_sizes
        start_poses = op.start_positions

        i_indices, j_indices = xp.mgrid[
            start_poses[0]: start_poses[0] + x.shape[0],
            start_poses[1]: start_poses[1] + x.shape[1]
        ]
        filtered = i_indices < j_indices
        i_indices, j_indices, x = \
            i_indices[filtered], j_indices[filtered], x[filtered]

        d = op.x_shape[0]
        row_sizes = xp.arange(d - 1, -1, -1)
        row_cum_sizes = xp.empty((d + 1,), dtype=int)
        row_cum_sizes[0] = 0
        xp.cumsum(row_sizes, out=row_cum_sizes[1:])
        to_indices = row_cum_sizes[i_indices] + j_indices - (d - row_sizes[i_indices])

        cum_chunk_size = [0] + np.cumsum(out_chunk_size).tolist()
        for i in range(len(out_chunk_size[0])):
            index_range = cum_chunk_size[i], cum_chunk_size[i + 1]
            filtered = (to_indices >= index_range[0]) & (to_indices < index_range[1])
            out_indices = to_indices[filtered] - cum_chunk_size[i]
            ctx[op.outputs[0].key, str(i)] = out_indices, x[filtered]

    @classmethod
    def _execute_map(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)

        if len(inputs) == 2 and not inputs[1]:
            # check fail
            raise ValueError('Distance matrix X must be symmetric.')

        if xp is cp:  # pragma: no cover
            raise NotImplementedError('`squareform` does not support running on GPU yet')

        with device(device_id):
            x = inputs[0]
            if x.ndim == 1:
                cls._to_matrix(ctx, xp, x, op)
            else:
                cls._to_vector(ctx, xp, x, op)

    @classmethod
    def _execute_reduce(cls, ctx, op):
        input_keys, _ = get_shuffle_input_keys_idxes(op.inputs[0])
        raw_inputs = [ctx[(input_key, op.shuffle_key)] for input_key in input_keys]
        raw_indices = [inp[0] for inp in raw_inputs]
        raw_dists = [inp[1] for inp in raw_inputs]
        inputs, device_id, xp = as_same_device(
            raw_indices + raw_dists, op.device, ret_extra=True)
        raw_indices = inputs[:len(raw_indices)]
        raw_dists = inputs[len(raw_indices):]
        output = op.outputs[0]

        with device(device_id):
            out_dists = xp.zeros(output.shape, dtype=output.dtype)
            indices = xp.concatenate(raw_indices)
            dists = xp.concatenate(raw_dists)
            out_dists.flat[indices] = dists
            ctx[output.key] = out_dists

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls._execute_reduce(ctx, op)
        else:
            from scipy.spatial.distance import squareform

            (x,), device_id, xp = as_same_device(
                [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)

            if xp is cp:  # pragma: no cover
                raise NotImplementedError('`squareform` does not support running on GPU yet')

            with device(device_id):
                ctx[op.outputs[0].key] = squareform(x, checks=op.checks)


@require_module('scipy.spatial.distance')
def squareform(X, force='no', checks=True, chunk_size=None):
    """
    Convert a vector-form distance vector to a square-form distance
    matrix, and vice-versa.

    Parameters
    ----------
    X : Tensor
        Either a condensed or redundant distance matrix.
    force : str, optional
        As with MATLAB(TM), if force is equal to ``'tovector'`` or
        ``'tomatrix'``, the input will be treated as a distance matrix or
        distance vector respectively.
    checks : bool, optional
        If set to False, no checks will be made for matrix
        symmetry nor zero diagonals. This is useful if it is known that
        ``X - X.T1`` is small and ``diag(X)`` is close to zero.
        These values are ignored any way so they do not disrupt the
        squareform transformation.

    Returns
    -------
    Y : Tensor
        If a condensed distance matrix is passed, a redundant one is
        returned, or if a redundant one is passed, a condensed distance
        matrix is returned.

    Notes
    -----
    1. v = squareform(X)

       Given a square d-by-d symmetric distance matrix X,
       ``v = squareform(X)`` returns a ``d * (d-1) / 2`` (or
       :math:`{n \\choose 2}`) sized vector v.

      :math:`v[{n \\choose 2}-{n-i \\choose 2} + (j-i-1)]` is the distance
      between points i and j. If X is non-square or asymmetric, an error
      is returned.

    2. X = squareform(v)

      Given a ``d*(d-1)/2`` sized v for some integer ``d >= 2`` encoding
      distances as described, ``X = squareform(v)`` returns a d by d distance
      matrix X.  The ``X[i, j]`` and ``X[j, i]`` values are set to
      :math:`v[{n \\choose 2}-{n-i \\choose 2} + (j-i-1)]` and all
      diagonal elements are zero.

    """

    X = ascontiguousarray(X)

    op = TensorSquareform(checks=checks, dtype=X.dtype, gpu=X.op.gpu)
    return op(X, force=force, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..._utils cimport to_str


cpdef bytes new_actor_id():
    return np.random.bytes(32)


def create_actor_ref(*args, **kwargs):
    from ..core import ActorRef

    cdef str address
    cdef object uid

    address = to_str(kwargs.pop('address', None))
    uid = kwargs.pop('uid', None)

    if kwargs:
        raise ValueError('Only `address` or `uid` keywords are supported')

    if len(args) == 2:
        if address:
            raise ValueError('address has been specified')
        address = to_str(args[0])
        uid = args[1]
    elif len(args) == 1 and isinstance(args[0], ActorRef):
        uid = args[0].uid
        address = to_str(address or args[0].address)
    elif len(args) == 1:
        uid = args[0]

    if uid is None:
        raise ValueError('Actor uid should be provided')

    return ActorRef(address, uid)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import shutil
from typing import List, Dict, Union, Tuple, Iterator, BinaryIO, TextIO

from ...utils import implements, stringify_path
from .base import FileSystem, path_type


class LocalFileSystem(FileSystem):

    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = LocalFileSystem()
        return cls._instance

    @implements(FileSystem.cat)
    def cat(self, path: path_type):
        with self.open(path, 'rb') as f:
            return f.read()

    @implements(FileSystem.ls)
    def ls(self, path: path_type) -> List[path_type]:
        path = stringify_path(path)
        return sorted(os.path.join(path, x) for x in os.listdir(path))

    @implements(FileSystem.delete)
    def delete(self,
               path: path_type,
               recursive: bool = False):
        if os.path.isfile(path):
            os.remove(path)
        elif not recursive:
            os.rmdir(path)
        else:
            shutil.rmtree(path)

    @implements(FileSystem.rename)
    def rename(self,
               path: path_type,
               new_path: path_type):
        os.rename(path, new_path)

    @implements(FileSystem.stat)
    def stat(self, path: path_type) -> Dict:
        os_stat = os.stat(path)
        stat = dict(name=path, size=os_stat.st_size,
                    modified_time=os_stat.st_mtime)
        if os.path.isfile(path):
            stat['type'] = 'file'
        elif os.path.isdir(path):
            stat['type'] = 'directory'
        else:  # pragma: no cover
            stat['type'] = 'other'
        return stat

    @implements(FileSystem.mkdir)
    def mkdir(self,
              path: path_type,
              create_parents: bool = True):
        path = stringify_path(path)
        if create_parents:
            os.makedirs(path)
        else:
            os.mkdir(path)

    @implements(FileSystem.isdir)
    def isdir(self, path: path_type) -> bool:
        path = stringify_path(path)
        return os.path.isdir(path)

    @implements(FileSystem.isfile)
    def isfile(self, path: path_type) -> bool:
        path = stringify_path(path)
        return os.path.isfile(path)

    @implements(FileSystem._isfilestore)
    def _isfilestore(self) -> bool:
        return True

    @implements(FileSystem.exists)
    def exists(self, path: path_type):
        path = stringify_path(path)
        return os.path.exists(path)

    @implements(FileSystem.open)
    def open(self,
             path: path_type,
             mode: str = 'rb') -> Union[BinaryIO, TextIO]:
        path = stringify_path(path)
        return open(path, mode=mode)

    @implements(FileSystem.walk)
    def walk(self, path: path_type) -> Iterator[Tuple[str, List[str], List[str]]]:
        path = stringify_path(path)
        return os.walk(path)

    @implements(FileSystem.glob)
    def glob(self,
             path: path_type,
             recursive: bool = False) -> List[path_type]:
        path = stringify_path(path)
        return glob.glob(path, recursive=recursive)

    @property
    def pathsep(self) -> str:
        return os.path.sep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .datasource import tensor, named_tensor, array, asarray, ascontiguousarray, asfortranarray, \
    scalar, empty, empty_like, ones, ones_like, zeros, zeros_like, \
    full, full_like, arange, diag, diagflat, eye, identity, linspace, \
    meshgrid, indices, tril, triu, fromtiledb, fromtiledb as from_tiledb, from_dataframe, \
    fromhdf5, fromhdf5 as from_hdf5, fromzarr, fromzarr as from_zarr, from_vineyard
from .datastore import totiledb, totiledb as to_tiledb, tohdf5, tohdf5 as to_hdf5, \
    tozarr, tozarr as to_zarr, tovineyard, tovineyard as to_vineyard  # pylint: disable=reimported
from .base import result_type, ndim, copyto, transpose, where, broadcast_to, broadcast_arrays, \
    expand_dims, rollaxis, swapaxes, moveaxis, ravel, atleast_1d, atleast_2d, atleast_3d, argwhere, \
    array_split, split, hsplit, vsplit, dsplit, roll, squeeze, diff, ediff1d, \
    flip, flipud, fliplr, repeat, tile, isin, searchsorted, unique, \
    sort, argsort, partition, argpartition, topk, argtopk, copy, trapz, shape, \
    insert, delete
from .arithmetic import add, subtract, multiply, divide, truediv as true_divide, \
    floordiv as floor_divide, mod, power, float_power, fmod, sqrt, \
    around, round_, round_ as round, logaddexp, logaddexp2, negative, positive, \
    absolute, fabs, absolute as abs, rint, sign, degrees, radians, conj, conjugate, exp, exp2, \
    log, log2, log10, expm1, log1p, square, cbrt, reciprocal, \
    equal, not_equal, less, less_equal, greater, greater_equal, sin, cos, tan, \
    arcsin, arccos, arctan, arctan2, hypot, sinh, cosh, tanh, arcsinh, arccosh, arctanh, \
    deg2rad, rad2deg, bitand as bitwise_and, bitor as bitwise_or, bitxor as bitwise_xor, \
    invert, invert as bitwise_not, lshift as left_shift, rshift as right_shift, \
    logical_and, logical_or, logical_xor, logical_not, \
    maximum, minimum, floor, ceil, trunc, remainder, fmax, fmin, isfinite, isinf, isnan, \
    signbit, copysign, nextafter, spacing, clip, isclose, ldexp, frexp, modf, angle, \
    isreal, iscomplex, real, imag, fix, i0, sinc, nan_to_num, tree_add, tree_multiply
from .statistics import average, cov, corrcoef, digitize, ptp, \
    histogram_bin_edges, histogram, quantile, percentile, median
from .linalg.tensordot import tensordot
from .linalg.dot import dot
from .linalg.inner import inner, innerproduct
from .linalg.vdot import vdot
from .linalg.matmul import matmul
from .reduction import sum, nansum, prod, prod as product, nanprod, \
    max, max as amax, nanmax, min, min as amin, nanmin, all, any, mean, nanmean, \
    argmax, nanargmax, argmin, nanargmin, cumsum, cumprod, \
    var, std, nanvar, nanstd, nancumsum, nancumprod, count_nonzero, allclose, array_equal
from .reshape import reshape
from .merge import concatenate, stack, hstack, vstack, dstack, column_stack, union1d, block
from .indexing import take, compress, extract, choose, unravel_index, \
    nonzero, flatnonzero, fill_diagonal
from .rechunk import rechunk
from .einsum import einsum
from .images import imread
# noinspection PyUnresolvedReferences
from .lib.index_tricks import mgrid, ogrid, ndindex, r_, c_
from .core import mutable_tensor

from . import random
from . import fft
from . import linalg
from . import lib
from . import special
from . import stats

# types
from .core import Tensor, MutableTensor
# noinspection PyUnresolvedReferences
from ..core import ExecutableTuple

# noinspection PyUnresolvedReferences
from numpy import newaxis, AxisError, inf, Inf, NINF, nan, NAN, NaN, pi, e, \
    errstate, geterr, seterr
# import numpy types
# noinspection PyUnresolvedReferences
from numpy import dtype, object, number, inexact, floating, complexfloating, \
    integer, signedinteger, unsignedinteger, character, generic, flexible, \
    int, int_, bool, bool_, float_, cfloat, bytes_, unicode_, void, object_, \
    intc, intp, int8, int16, int32, int64, uint8, uint16, uint32, uint64, uint, \
    float, float16, float32, float64, double, complex64, complex128, datetime64, timedelta64
# noinspection PyUnresolvedReferences
from numpy import finfo

# register fuse op and fetch op
from .fuse import TensorFuseChunk, TensorCpFuseChunk, TensorNeFuseChunk
from .fetch import TensorFetch, TensorFetchShuffle
from . import ufunc
del TensorFuseChunk, TensorCpFuseChunk, TensorNeFuseChunk, TensorFetch, TensorFetchShuffle, ufunc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
import operator
from functools import partial, reduce
from collections.abc import Iterable

from .array import SparseNDArray, call_sparse
from .matrix import SparseMatrix
from .vector import SparseVector
from .core import issparse, get_sparse_module

from .coo import COONDArray


def asarray(x, shape=None):
    from .core import issparse

    if issparse(x):
        return SparseNDArray(x, shape=shape)

    return x


def add(a, b, **_):
    try:
        return a + b
    except TypeError:
        if hasattr(b, '__radd__'):
            return b.__radd__(a)
        raise


def subtract(a, b, **_):
    try:
        return a - b
    except TypeError:
        if hasattr(b, '__rsub__'):
            return b.__rsub__(a)
        raise


def multiply(a, b, **_):
    try:
        return a * b
    except TypeError:
        if hasattr(b, '__rmul__'):
            return b.__rmul__(a)
        raise


def divide(a, b, **_):
    try:
        return a / b
    except TypeError:
        if hasattr(b, '__rdiv__'):
            return b.__rdiv__(a)
        raise


def true_divide(a, b, **_):
    try:
        return a / b
    except TypeError:
        if hasattr(b, '__rtruediv__'):
            return b.__rtruediv__(a)
        raise


def floor_divide(a, b, **_):
    try:
        return a // b
    except TypeError:
        if hasattr(b, '__rfloordiv__'):
            return b.__rfloordiv__(a)
        raise


def power(a, b, **_):
    try:
        return a ** b
    except TypeError:
        if hasattr(b, '__rpow__'):
            return b.__rpow__(a)
        raise


def mod(a, b, **_):
    try:
        return a % b
    except TypeError:
        if hasattr(b, '__rmod__'):
            return b.__rmod__(a)
        raise


def _call_bin(method, a, b, **kwargs):
    from .core import get_array_module, cp, issparse

    # order does not take effect for sparse
    kwargs.pop('order', None)
    if hasattr(a, method):
        res = getattr(a, method)(b, **kwargs)
    elif get_array_module(a).isscalar(a):
        res = call_sparse(method, a, b, **kwargs)
    else:
        assert get_array_module(a) == get_array_module(b)
        xp = get_array_module(a)
        try:
            res = getattr(xp, method)(a, b, **kwargs)
        except TypeError:
            if xp is cp and issparse(b):
                res = getattr(xp, method)(a, b.toarray(), **kwargs)
            else:
                raise

    if res is NotImplemented:
        raise NotImplementedError

    return res


def _call_unary(method, x, *args, **kwargs):
    from .core import get_array_module

    # order does not take effect for sparse
    kwargs.pop('order', None)
    if hasattr(x, method):
        res = getattr(x, method)(*args, **kwargs)
    else:
        xp = get_array_module(x)
        res = getattr(xp, method)(x, *args, **kwargs)

    if res is NotImplemented:
        raise NotImplementedError

    return res


def float_power(a, b, **kw):
    return _call_bin('float_power', a, b, **kw)


def fmod(a, b, **kw):
    return _call_bin('fmod', a, b, **kw)


def logaddexp(a, b, **kw):
    return _call_bin('logaddexp', a, b, **kw)


def logaddexp2(a, b, **kw):
    return _call_bin('logaddexp2', a, b, **kw)


def negative(x, **_):
    return -x


def positive(x, **_):
    return operator.pos(x)


def absolute(x, **_):
    return builtins.abs(x)


abs = absolute


fabs = partial(_call_unary, 'fabs')


def rint(x, **kw):
    return _call_unary('rint', x, **kw)


def sign(x, **kw):
    return _call_unary('sign', x, **kw)


def conj(x, **kw):
    return _call_unary('conj', x, **kw)


def exp(x, **kw):
    return _call_unary('exp', x, **kw)


def exp2(x, **kw):
    return _call_unary('exp2', x, **kw)


def log(x, **kw):
    return _call_unary('log', x, **kw)


def log2(x, **kw):
    return _call_unary('log2', x, **kw)


def log10(x, **kw):
    return _call_unary('log10', x, **kw)


def expm1(x, **kw):
    return _call_unary('expm1', x, **kw)


def log1p(x, **kw):
    return _call_unary('log1p', x, **kw)


def sqrt(x, **kw):
    return _call_unary('sqrt', x, **kw)


def square(x, **kw):
    return _call_unary('square', x, **kw)


def cbrt(x, **kw):
    return _call_unary('cbrt', x, **kw)


def reciprocal(x, **kw):
    return _call_unary('reciprocal', x, **kw)


gamma = partial(_call_unary, 'gamma')
gammaln = partial(_call_unary, 'gammaln')
loggamma = partial(_call_unary, 'loggamma')
gammasgn = partial(_call_unary, 'gammasgn')
gammainc = partial(_call_bin, 'gammainc')
gammaincinv = partial(_call_bin, 'gammaincinv')
gammaincc = partial(_call_bin, 'gammaincc')
gammainccinv = partial(_call_bin, 'gammainccinv')
beta = partial(_call_bin, 'beta')
betaln = partial(_call_bin, 'betaln')
betainc = partial(call_sparse, 'betainc')
betaincinv = partial(call_sparse, 'betaincinv')
psi = partial(_call_unary, 'psi')
rgamma = partial(_call_unary, 'rgamma')
polygamma = partial(_call_bin, 'polygamma')
multigammaln = partial(_call_bin, 'multigammaln')
digamma = partial(_call_unary, 'digamma')
poch = partial(_call_bin, 'poch')

entr = partial(_call_unary, 'entr')
rel_entr = partial(_call_bin, 'rel_entr')
kl_div = partial(_call_bin, 'kl_div')

xlogy = partial(_call_bin, 'xlogy')

erf = partial(_call_unary, 'erf')


def equal(a, b, **_):
    try:
        return a == b
    except TypeError:
        return b == a


def not_equal(a, b, **_):
    try:
        return a != b
    except TypeError:
        return b != a


def less(a, b, **_):
    try:
        return a < b
    except TypeError:
        return b > a


def less_equal(a, b, **_):
    try:
        return a <= b
    except TypeError:
        return b >= a


def greater(a, b, **_):
    try:
        return a > b
    except TypeError:
        return b < a


def greater_equal(a, b, **_):
    try:
        return a >= b
    except TypeError:
        return b <= a


def logical_and(a, b, **kw):
    return _call_bin('logical_and', a, b, **kw)


def logical_or(a, b, **kw):
    return _call_bin('logical_or', a, b, **kw)


def logical_xor(a, b, **kw):
    return _call_bin('logical_xor', a, b, **kw)


def logical_not(x, **kw):
    return _call_unary('logical_not', x, **kw)


def isclose(a, b, **kw):
    return _call_bin('isclose', a, b, **kw)


def bitwise_and(a, b, **_):
    try:
        return a & b
    except TypeError:
        return b & a


def bitwise_or(a, b, **_):
    try:
        return a | b
    except TypeError:
        return b | a


def bitwise_xor(a, b, **_):
    try:
        return operator.xor(a, b)
    except TypeError:
        return operator.xor(b, a)


def invert(x, **_):
    return ~x


def left_shift(a, b, **_):
    return a << b


def right_shift(a, b, **_):
    return a >> b


def sin(x, **kw):
    return _call_unary('sin', x, **kw)


def cos(x, **kw):
    return _call_unary('cos', x, **kw)


def tan(x, **kw):
    return _call_unary('tan', x, **kw)


def arcsin(x, **kw):
    return _call_unary('arcsin', x, **kw)


def arccos(x, **kw):
    return _call_unary('arccos', x, **kw)


def arctan(x, **kw):
    return _call_unary('arctan', x, **kw)


def arctan2(a, b, **kw):
    return _call_bin('arctan2', a, b, **kw)


def hypot(a, b, **kw):
    return _call_bin('hypot', a, b, **kw)


def sinh(x, **kw):
    return _call_unary('sinh', x, **kw)


def cosh(x, **kw):
    return _call_unary('cosh', x, **kw)


def tanh(x, **kw):
    return _call_unary('tanh', x, **kw)


def arcsinh(x, **kw):
    return _call_unary('arcsinh', x, **kw)


def arccosh(x, **kw):
    return _call_unary('arccosh', x, **kw)


def around(x, **kw):
    return _call_unary('around', x, **kw)


def arctanh(x, **kw):
    return _call_unary('arctanh', x, **kw)


def deg2rad(x, **kw):
    return _call_unary('deg2rad', x, **kw)


def rad2deg(x, **kw):
    return _call_unary('rad2deg', x, **kw)


def angle(x, **kw):
    return _call_unary('angle', x, **kw)


def isinf(x, **kw):
    return _call_unary('isinf', x, **kw)


def isnan(x, **kw):
    return _call_unary('isnan', x, **kw)


def signbit(x, **kw):
    return _call_unary('signbit', x, **kw)


def dot(a, b, sparse=True, **_):
    from .core import issparse

    if not issparse(a):
        ret = a.dot(b)
        if not sparse:
            return ret
        else:
            xps = get_sparse_module(ret)
            return SparseNDArray(xps.csr_matrix(ret), shape=ret.shape)

    return a.dot(b, sparse=sparse)


def tensordot(a, b, axes=2, sparse=True):
    if isinstance(axes, Iterable):
        a_axes, b_axes = axes
    else:
        a_axes = tuple(range(a.ndim - 1, a.ndim - axes - 1, -1))
        b_axes = tuple(range(0, axes))

    if isinstance(a_axes, Iterable):
        a_axes = tuple(a_axes)
    else:
        a_axes = (a_axes,)
    if isinstance(b_axes, Iterable):
        b_axes = tuple(b_axes)
    else:
        b_axes = (b_axes,)

    if a_axes == (a.ndim - 1,) and b_axes == (b.ndim - 2,):
        return dot(a, b, sparse=sparse)

    if a.ndim == b.ndim == 2:
        if a_axes == (a.ndim - 1,) and b_axes == (b.ndim - 1,):
            # inner product of multiple dims
            return dot(a, b.T, sparse=sparse)

    if a.ndim == 1 or b.ndim == 1:
        return dot(a, b, sparse=sparse)

    raise NotImplementedError


def matmul(a, b, sparse=True, **_):
    return dot(a, b, sparse=sparse)


def concatenate(tensors, axis=0):
    return reduce(lambda a, b: _call_bin('concatenate', a, b, axis=axis), tensors)


def transpose(tensor, axes=None):
    return _call_unary('transpose', tensor, axes=axes)


def swapaxes(tensor, axis1, axis2):
    return _call_unary('swapaxes', tensor, axis1, axis2)


def sum(tensor, axis=None, **kw):
    return _call_unary('sum', tensor, axis=axis, **kw)


def prod(tensor, axis=None, **kw):
    return _call_unary('prod', tensor, axis=axis, **kw)


def amax(tensor, axis=None, **kw):
    return _call_unary('amax', tensor, axis=axis, **kw)


max = amax


def amin(tensor, axis=None, **kw):
    return _call_unary('amin', tensor, axis=axis, **kw)


min = amin


def all(tensor, axis=None, **kw):
    return _call_unary('all', tensor, axis=axis, **kw)


def any(tensor, axis=None, **kw):
    return _call_unary('any', tensor, axis=axis, **kw)


def mean(tensor, axis=None, **kw):
    return _call_unary('mean', tensor, axis=axis, **kw)


def nansum(tensor, axis=None, **kw):
    return _call_unary('nansum', tensor, axis=axis, **kw)


def nanprod(tensor, axis=None, **kw):
    return _call_unary('nanprod', tensor, axis=axis, **kw)


def nanmax(tensor, axis=None, **kw):
    return _call_unary('nanmax', tensor, axis=axis, **kw)


def nanmin(tensor, axis=None, **kw):
    return _call_unary('nanmin', tensor, axis=axis, **kw)


def argmax(tensor, axis=None, **kw):
    return _call_unary('argmax', tensor, axis=axis, **kw)


def nanargmax(tensor, axis=None, **kw):
    return _call_unary('nanargmax', tensor, axis=axis, **kw)


def argmin(tensor, axis=None, **kw):
    return _call_unary('argmin', tensor, axis=axis, **kw)


def nanargmin(tensor, axis=None, **kw):
    return _call_unary('nanargmin', tensor, axis=axis, **kw)


def var(tensor, axis=None, **kw):
    return _call_unary('var', tensor, axis=axis, **kw)


def cumsum(tensor, axis=None, **kw):
    return _call_unary('cumsum', tensor, axis=axis, **kw)


def cumprod(tensor, axis=None, **kw):
    return _call_unary('cumprod', tensor, axis=axis, **kw)


def nancumsum(tensor, axis=None, **kw):
    return _call_unary('nancumsum', tensor, axis=axis, **kw)


def nancumprod(tensor, axis=None, **kw):
    return _call_unary('nancumprod', tensor, axis=axis, **kw)


def count_nonzero(tensor, axis=None, **kw):
    return _call_unary('count_nonzero', tensor, axis=axis, **kw)


def maximum(a, b, **kw):
    return _call_bin('maximum', a, b, **kw)


def minimum(a, b, **kw):
    return _call_bin('minimum', a, b, **kw)


def fmax(a, b, **kw):
    return _call_bin('fmax', a, b, **kw)


def fmin(a, b, **kw):
    return _call_bin('fmin', a, b, **kw)


def floor(x, **kw):
    return _call_unary('floor', x, **kw)


def ceil(x, **kw):
    return _call_unary('ceil', x, **kw)


def trunc(x, **kw):
    return _call_unary('trunc', x, **kw)


def degrees(x, **kw):
    return _call_unary('degrees', x, **kw)


def radians(x, **kw):
    return _call_unary('radians', x, **kw)


def clip(a, a_max, a_min, **kw):
    from .core import get_array_module

    if hasattr(a, 'clip'):
        res = getattr(a, 'clip')(a_max, a_min, **kw)
    else:
        xp = get_array_module(a)
        res = getattr(xp, 'clip')(a, a_max, a_min, **kw)

    if res is NotImplemented:
        raise NotImplementedError

    return res


def iscomplex(x, **kw):
    return _call_unary('iscomplex', x, **kw)


def real(x, **_):
    return x.real


def imag(x, **_):
    return x.imag


def fix(x, **kw):
    return _call_unary('fix', x, **kw)


def i0(x, **kw):
    return _call_unary('i0', x, **kw)


def nan_to_num(x, **kw):
    return _call_unary('nan_to_num', x, **kw)


def copysign(a, b, **kw):
    return _call_bin('copysign', a, b, **kw)


def nextafter(a, b, **kw):
    return _call_bin('nextafter', a, b, **kw)


def spacing(x, **kw):
    return _call_unary('spacing', x, **kw)


def ldexp(a, b, **kw):
    return _call_bin('ldexp', a, b, **kw)


def frexp(x, **kw):
    return _call_unary('frexp', x, **kw)


def modf(x, **kw):
    return _call_unary('modf', x, **kw)


def sinc(x, **kw):
    return _call_unary('sinc', x, **kw)


def isfinite(x, **kw):
    return _call_unary('isfinite', x, **kw)


def isreal(x, **kw):
    return _call_unary('isreal', x, **kw)


def where(cond, x, y):
    if any([i.ndim not in (0, 2) for i in (cond, x, y)]):
        raise NotImplementedError

    from .matrix import where as matrix_where
    return matrix_where(cond, x, y)


def digitize(x, bins, right=False):
    return _call_unary('digitize', x, bins, right)


def repeat(a, repeats, axis=None):
    return _call_unary('repeat', a, repeats, axis=axis)


def fill_diagonal(a, val, wrap=False):
    return _call_unary('fill_diagonal', a, val, wrap=wrap)


def unique(a, return_index=False, return_inverse=False,
           return_counts=False, axis=None):
    return _call_unary('unique', a, return_index=return_index,
                       return_inverse=return_inverse,
                       return_counts=return_counts,
                       axis=axis)


def zeros(shape, dtype=float, gpu=False):
    if len(shape) == 2:
        from .matrix import zeros_sparse_matrix

        return zeros_sparse_matrix(shape, dtype=dtype, gpu=gpu)

    raise NotImplementedError


def ones_like(x):
    from .core import get_array_module

    return get_array_module(x).ones(x.shape)


def diag(v, k=0, gpu=False):
    assert v.ndim in {1, 2}

    from .matrix import diag_sparse_matrix

    return diag_sparse_matrix(v, k=k, gpu=gpu)


def eye(N, M=None, k=0, dtype=float, gpu=False):
    from .matrix import eye_sparse_matrix

    return eye_sparse_matrix(N, M=M, k=k, dtype=dtype, gpu=gpu)


def triu(m, k=0, gpu=False):
    if m.ndim == 2:
        from .matrix import triu_sparse_matrix

        return triu_sparse_matrix(m, k=k, gpu=gpu)

    raise NotImplementedError


def tril(m, k=0, gpu=False):
    if m.ndim == 2:
        from .matrix import tril_sparse_matrix

        return tril_sparse_matrix(m, k=k, gpu=gpu)

    raise NotImplementedError


def lu(m):
    from .matrix import lu_sparse_matrix

    return lu_sparse_matrix(m)


def solve_triangular(a, b, lower=False, sparse=True):
    from .matrix import solve_triangular_sparse_matrix

    return solve_triangular_sparse_matrix(a, b, lower=lower, sparse=sparse)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from .core import DataFrameDeviceConversionBase


class DataFrameToCPU(DataFrameDeviceConversionBase):
    _op_type_ = OperandDef.TO_CPU

    def __init__(self, dtypes=None, gpu=None, sparse=None, output_types=None, **kw):
        super().__init__(_dtypes=dtypes, _gpu=gpu, _sparse=sparse, _output_types=output_types, **kw)
        if self._gpu or self._gpu is None:
            self._gpu = False

    @classmethod
    def execute(cls, ctx, op):
        ctx[op.outputs[0].key] = ctx[op.inputs[0].key].to_pandas()


def to_cpu(df_or_series):
    if df_or_series.op.gpu is False:
        # if op.gpu is None, means unknown
        return df_or_series

    op = DataFrameToCPU()
    return op(df_or_series)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import logging
import struct
import sys
import threading
import weakref

import numpy as np

from .actors import FunctionActor
from .actors.core import ActorRef
from .errors import PromiseTimeout
from .utils import build_exc_info

logger = logging.getLogger(__name__)
_promise_pool = dict()
_pack_qword = struct.Struct('<Q').pack


class Promise(object):
    """
    Object representing a promised result
    """
    def __init__(self, resolve=None, reject=None, done=False, failed=False,
                 args=None, kwargs=None):
        # use random promise id
        self._id = _pack_qword(id(self)) + np.random.bytes(32)

        self._accept_handler = self._wrap_handler(resolve)
        self._reject_handler = self._wrap_handler(reject)

        # _bind_item indicates the Promise object whose step_next()
        # should be called when current Promise finishes. For instance,
        # in ``p2 = p1.then(lambda arg: promise_call())``, the value of
        # _bind_item of the Promise object returned by ``promise_call()``
        # is p2, thus when the lambda finishes, subsequent operations in
        # p2 can be executed.
        self._bind_item = None  # type: Promise
        # _next_item indicates the next Promise object to be invoked
        # when the current one returns.
        self._next_item = None  # type: Promise

        # promise results
        if done:
            self._accepted = True
        elif failed:
            self._accepted = False
        else:
            self._accepted = None
            # register in global pool to reject gc collection when calls not finished
            _promise_pool[self._id] = self

        self._args = args or ()
        self._kwargs = kwargs or {}

        self.post_create()

    def __del__(self):
        self.pre_destroy()
        self._clear_result_cache()

    @property
    def id(self):
        return self._id

    def post_create(self):
        pass

    def pre_destroy(self):
        pass

    def _wrap_handler(self, func):
        """
        Wraps a promise handler
        """
        if func is None:
            return None

        @functools.wraps(func)
        def _wrapped(*args, **kwargs):
            _promise_pool.pop(self.id, None)

            result = None
            try:
                result = func(*args, **kwargs)
                if isinstance(result, Promise):
                    # the function itself returns a promise object
                    # bind returned promise object to current promise
                    assert result._next_item is None
                    result._bind_item = self
                    if result._accepted is not None:
                        # promise already done, we move next
                        args = result._args or ()
                        kwargs = result._kwargs or {}
                        result._clear_result_cache()

                        kwargs['_accept'] = result._accepted
                        return result._internal_step_next, args, kwargs
                else:
                    # return non-promise result, we just step next
                    return self._internal_step_next, (result,), {}
            except:  # noqa: E722
                # error occurred when executing func, we reject with exc_info
                logger.exception('Exception met in executing promise.')
                exc = sys.exc_info()
                self._clear_result_cache()
                return self._internal_step_next, exc, dict(_accept=False)
            finally:
                del result
        return _wrapped

    def _get_bind_root(self):
        """
        Get root promise of result promises
        :return: root promise
        :rtype: Promise
        """
        target = self
        while target._bind_item is not None:
            if target.id in _promise_pool:
                # remove binder that will not be used later
                _promise_pool.pop(target.id, None)
            target._clear_result_cache()
            target = target._bind_item
        return target

    @staticmethod
    def _get_handling_promise(p, handler_attr):
        """
        Get a promise object that defines the handler needed
        :param Promise p: promise object
        :param str handler_attr: attribute name of the handler
        :rtype: Promise
        """
        while getattr(p, handler_attr) is None:
            p._accept_handler = p._reject_handler = None

            p = p._get_bind_root()  # type: Promise
            if p and p._next_item is not None:
                _promise_pool.pop(p.id, None)
                p._clear_result_cache()
                p = p._next_item
            else:
                break
        return p

    @staticmethod
    def _log_unexpected_error(args):
        if args and len(args) == 3 and issubclass(args[0], Exception):
            logger.exception('Unhandled exception in promise', exc_info=args)

    def _clear_result_cache(self):
        self._args = ()
        self._kwargs = {}

    def _internal_step_next(self, *args, **kwargs):
        """
        Actual call to step promise call into the next step.
        If there are further steps, the function will return it
        as a (func, args, kwargs) tuple to reduce memory footprint.
        """
        accept = kwargs.pop('_accept', True)
        target_promise = self

        self._accepted = accept

        try:
            root_promise = self._get_bind_root()

            if root_promise:
                _promise_pool.pop(root_promise.id, None)

            target_promise = root_promise._next_item
            root_promise._accepted = accept
            if not target_promise:
                root_promise._args = args
                root_promise._kwargs = kwargs
                if not accept:
                    self._log_unexpected_error(args)
                return
            else:
                root_promise._clear_result_cache()

            next_call = None
            if accept:
                acceptor = self._get_handling_promise(target_promise, '_accept_handler')
                if acceptor and acceptor._accept_handler:
                    # remove the handler in promise in case that
                    # function closure is not freed
                    handler, acceptor._accept_handler, acceptor._reject_handler = \
                        acceptor._accept_handler, None, None
                    next_call = handler(*args, **kwargs)
                else:
                    acceptor._accepted = accept
                    acceptor._args = args
                    acceptor._kwargs = kwargs
                    _promise_pool.pop(acceptor.id, None)
            else:
                rejecter = self._get_handling_promise(target_promise, '_reject_handler')
                if rejecter and rejecter._reject_handler:
                    # remove the handler in promise in case that
                    # function closure is not freed
                    handler, rejecter._reject_handler, rejecter._accept_handler = \
                        rejecter._reject_handler, None, None
                    next_call = handler(*args, **kwargs)
                else:
                    rejecter._accepted = accept
                    rejecter._args = args
                    rejecter._kwargs = kwargs
                    _promise_pool.pop(rejecter.id, None)
                    self._log_unexpected_error(args)
            return next_call
        finally:
            del args, kwargs
            if target_promise:
                _promise_pool.pop(target_promise.id, None)

    def step_next(self, args_and_kwargs=None):
        """
        Step into next promise with given args and kwargs
        """
        if args_and_kwargs is None:
            args_and_kwargs = [(), {}]
        step_call = (self._internal_step_next,) + tuple(args_and_kwargs)
        del args_and_kwargs[:]

        while step_call is not None:
            func, args, kwargs = step_call
            step_call = func(*args, **kwargs)

    def then(self, on_fulfilled, on_rejected=None):
        promise = Promise(on_fulfilled, on_rejected)
        assert self._bind_item is None
        self._next_item = promise
        if self._accepted is not None:
            args_and_kwargs = [self._args, self._kwargs]
            args_and_kwargs[1]['_accept'] = self._accepted
            _promise_pool.pop(self.id, None)
            self._clear_result_cache()
            self.step_next(args_and_kwargs)
        return promise

    def catch(self, on_rejected):
        return self.then(None, on_rejected)

    def wait(self, waiter=None, timeout=None):
        """
        Wait when the promise returns. Currently only used in debug.
        :param waiter: wait object
        :param timeout: wait timeout
        :return: accept or reject
        """
        waiter = threading.Event()
        status = []

        def _finish_exec(accept_or_reject):
            waiter.set()
            status.append(accept_or_reject)

        self.then(lambda *_, **__: _finish_exec(True),
                  lambda *_, **__: _finish_exec(False))
        waiter.wait(timeout)
        return status[0]


class PromiseRefWrapper(object):
    """
    Promise wrapper that enables promise call by adding _promise=True
    """
    def __init__(self, ref, caller):
        self._ref = caller.ctx.actor_ref(ref)
        self._caller = caller  # type: PromiseActor

    def send(self, message):
        return self._ref.send(message)

    def tell(self, message, delay=None):
        return self._ref.tell(message, delay=delay)

    def destroy(self):
        return self._ref.destroy_actor(self)

    @property
    def uid(self):
        return self._ref.uid

    @property
    def address(self):
        return self._ref.address

    def __reduce__(self):
        # when pickled, return original ActorRef
        return ActorRef, (self._caller.address, self._ref.uid)

    def __getattr__(self, item):
        if item.startswith('_'):
            return object.__getattribute__(self, item)

        def _mt_fun(*args, **kwargs):
            ref_fun = getattr(self._ref, item)
            if not kwargs.pop('_promise', False):
                return ref_fun(*args, **kwargs)

            timeout = kwargs.pop('_timeout', 0)
            spawn = kwargs.pop('_spawn', True)

            p = Promise()
            promise_id = p.id

            if timeout and timeout > 0:
                # add a callback that triggers some times later to deal with timeout
                timeout_coro = self._caller.ref().handle_promise_timeout(
                    promise_id, _tell=True, _delay=timeout)
            else:
                timeout_coro = None

            self._caller.register_promise(p, self._ref, timeout_coro)

            kwargs['callback'] = ((self._caller.uid, self._caller.address),
                                  'handle_promise', promise_id)
            kwargs['_tell'] = True

            def _promise_runner():
                try:
                    ref_fun(*args, **kwargs)
                except:  # noqa: E722
                    self._caller.ref().handle_promise(
                        promise_id, *sys.exc_info(), _accept=False, _tell=True)

            if spawn:
                self._caller.async_group.spawn(_promise_runner)
            else:
                ref_fun(*args, **kwargs)

            return p

        return _mt_fun


def reject_on_exception(func):
    """
    Decorator on actor callback functions that handles exceptions by
    sending it to caller as promise rejections. The function should have
    an argument called ``callback``.
    """
    arg_names = inspect.getfullargspec(func).args
    callback_pos = None
    if arg_names:
        for idx, name in enumerate(arg_names):
            if name == 'callback':
                callback_pos = idx
                break

    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        callback = None
        if 'callback' in kwargs:
            callback = kwargs['callback']
        elif callback_pos and callback_pos < len(args):
            callback = args[callback_pos]

        try:
            return func(*args, **kwargs)
        except:  # noqa: E722
            actor = args[0]
            logger.exception('Unhandled exception in promise call')
            if callback:
                actor.tell_promise(callback, *sys.exc_info(), _accept=False)
            else:
                raise
    return _wrapped


class PromiseActor(FunctionActor):
    """
    Actor class providing promise functionality
    """
    def _prepare_promise_registration(self):
        if not hasattr(self, '_promises'):
            self._promises = dict()
            self._promise_ref_keys = dict()
            self._promise_timeout_coros = dict()
            self._ref_key_promises = dict()

    def promise_ref(self, *args, **kwargs):
        """
        Wraps an existing ActorRef into a promise ref
        """
        self._prepare_promise_registration()

        if not args and not kwargs:
            ref = self.ref()
        elif args and isinstance(args[0], ActorRef):
            ref = self.ctx.actor_ref(args[0].uid, address=args[0].address)
        else:
            ref = self.ctx.actor_ref(*args, **kwargs)
        return PromiseRefWrapper(ref, self)

    @property
    def async_group(self):
        if not hasattr(self, '_async_group'):
            self._async_group = self.ctx.asyncpool()
        return self._async_group

    def spawn_promised(self, func, *args, **kwargs):
        """
        Run func asynchronously in a pool and returns a promise.
        The running of the function does not block current actor.

        :param func: function to run
        :return: promise
        """
        self._prepare_promise_registration()

        p = Promise()
        ref = self.ref()
        self.register_promise(p, ref)

        def _wrapped(*a, **kw):
            try:
                result = func(*a, **kw)
            except:  # noqa: E722
                ref.handle_promise(p.id, *sys.exc_info(), _accept=False, _tell=True)
            else:
                ref.handle_promise(p.id, result, _tell=True)
            finally:
                del a, kw

        try:
            self.async_group.spawn(_wrapped, *args, **kwargs)
        finally:
            del args, kwargs
        return p

    def register_promise(self, promise, ref, timeout_coro=None):
        """
        Register a promise into the actor with referrer info

        :param Promise promise: promise object to register
        :param ActorRef ref: ref
        :param timeout_coro: coroutine of timeout
        """
        promise_id = promise.id

        def _weak_callback(*_):
            self.delete_promise(promise_id)

        self._promises[promise_id] = weakref.ref(promise, _weak_callback)
        ref_key = (ref.uid, ref.address)
        self._promise_ref_keys[promise_id] = ref_key
        if timeout_coro is not None:
            self._promise_timeout_coros[promise_id] = timeout_coro
        try:
            self._ref_key_promises[ref_key].add(promise_id)
        except KeyError:
            self._ref_key_promises[ref_key] = {promise_id}

    def get_promise(self, promise_id):
        """
        Get promise object from weakref.
        """
        obj = self._promises.get(promise_id)
        if obj is None:
            return None
        return obj()

    def delete_promise(self, promise_id):
        _promise_pool.pop(promise_id, None)
        if promise_id not in self._promises:
            return
        ref_key = self._promise_ref_keys[promise_id]
        self._ref_key_promises[ref_key].remove(promise_id)
        del self._promises[promise_id]
        del self._promise_ref_keys[promise_id]

    def reject_dead_endpoints(self, dead_endpoints, *args, **kwargs):
        """
        Reject all promises related to given remote address
        :param dead_endpoints: list of dead workers
        """
        dead_refs = []
        for ref_key in self._ref_key_promises.keys():
            uid, addr = ref_key
            if addr in dead_endpoints:
                dead_refs.append(self.ctx.actor_ref(uid, address=addr))
        return self.reject_promise_refs(dead_refs, *args, **kwargs)

    def reject_promise_refs(self, refs, *args, **kwargs):
        """
        Reject all promises related to given actor ref
        :param refs: actor refs to reject
        """
        kwargs['_accept'] = False
        handled_refs = []
        for ref in refs:
            ref_key = (ref.uid, ref.address)
            if ref_key not in self._ref_key_promises:
                continue
            handled_refs.append(ref)
            for promise_id in list(self._ref_key_promises[ref_key]):
                p = self.get_promise(promise_id)
                if p is None:
                    continue
                p.step_next([args, kwargs])
        return handled_refs

    def tell_promise(self, callback, *args, **kwargs):
        """
        Tell promise results to the caller
        :param callback: promise callback
        """
        wait = kwargs.pop('_wait', False)
        uid, address = callback[0]
        callback_args = callback[1:] + args + (kwargs, )
        return self.ctx.actor_ref(uid, address=address).tell(callback_args, wait=wait)

    def handle_promise(self, promise_id, *args, **kwargs):
        """
        Callback entry for promise results
        :param promise_id: promise key
        """
        try:
            p = self.get_promise(promise_id)
            if p is None:
                logger.warning('Promise %r reentered or not registered in %s', promise_id, self.uid)
                return
            self.get_promise(promise_id).step_next([args, kwargs])

            timeout_coro = self._promise_timeout_coros.pop(promise_id, None)
            if timeout_coro is not None:
                timeout_coro.kill()
        finally:
            self.delete_promise(promise_id)

    def handle_promise_timeout(self, promise_id):
        """
        Callback entry for promise timeout
        :param promise_id: promise key
        """
        try:
            p = self.get_promise(promise_id)
            if p is None or p._accepted is not None:
                # skip promises that are already finished
                return

            p.step_next([build_exc_info(PromiseTimeout), dict(_accept=False)])
            self._promise_timeout_coros.pop(promise_id, None)
        finally:
            self.delete_promise(promise_id)


def all_(promises):
    """
    Create a promise with promises. Invoked when all referenced promises accepted
    or at least one rejected
    :param promises: collection of promises
    :return: the new promise
    """
    promises = list(promises)

    promise_refs = [weakref.ref(p) for p in promises if isinstance(p, Promise)]
    new_promise = Promise()
    running_ids = set(ref().id for ref in promise_refs)

    def _build_then_reject(promise_ref):
        ref_id = promise_ref().id

        def _then(*_, **__):
            _promise_pool.pop(ref_id, None)
            running_ids.difference_update([ref_id])
            if not running_ids:
                new_promise.step_next()

        def _reject(*args, **kw):
            _promise_pool.pop(ref_id, None)
            if new_promise._accepted is not None:
                return
            kw['_accept'] = False
            new_promise.step_next([args, kw])

        return _then, _reject

    for ref in promise_refs:
        ref().then(*_build_then_reject(ref))

    if promise_refs:
        return new_promise
    else:
        new_promise.step_next()
        return new_promise


def get_active_promise_count():
    return len(_promise_pool)


def finished(*args, **kwargs):
    """
    Create a promise already finished
    :return: Promise object if _promise is True, otherwise args[0] will be returned
    """
    promise = kwargs.pop('_promise', True)
    accept = kwargs.pop('_accept', True)
    if not promise:
        return args[0] if args else None
    if accept:
        return Promise(done=True, args=args, kwargs=kwargs)
    else:
        return Promise(failed=True, args=args, kwargs=kwargs)
